#include <iostream>
#include <sstream>

#include <sqlite3.h>
#include <glibmm.h>
#include <boost/variant.hpp>
#include <boost/tuple/tuple.hpp>
#include "database.hpp"

namespace {

  static const char* sql_valuenames[] =
  {
      "INTEGER",    //SQLite doesn't know a boolean type so we have to convert hence and forth
      "INTEGER",
      "REAL",
      "TEXT"
  };

};

namespace Bmp {

      namespace DB {

	  void
	  DB::sqlite_update_hook (void*		data,
				  int		mod,
				  char const*   database,
				  char const*   table,
				  sqlite_int64  rowid)		      
	  {
	    Bmp::DB::DB *db = reinterpret_cast<Bmp::DB::DB *>(data);

	    HookType type;

	    switch (mod) 
	    {
	      case SQLITE_INSERT: type = HOOK_INSERT; break;
	      case SQLITE_DELETE: type = HOOK_DELETE; break;
	      case SQLITE_UPDATE: type = HOOK_UPDATE; break;
	      default: return;
	    }

	    TableHookMap::iterator hooks_i = db->hooks[type].find (table);
	    if (hooks_i == db->hooks[type].end()) return;
	    Hooks& hooks_ = (*hooks_i).second;
	    if (hooks_.empty()) return;

	    for (Hooks::iterator _slot = hooks_.begin (), end = hooks_.end (); _slot != end; ++_slot)
	    {
	      SlotSQLHook& hook = (*_slot);
	      hook (rowid);
	    }
	  }

	  //Creating hooks  
	  void
	  DB::add_hook	      (const std::string& name,
			       HookType		  type,
                               const SlotSQLHook& slot)
	  {
	    TableHookMap::iterator hooks_i = hooks[type].find (name);
	    if (hooks_i == hooks[type].end ())
	    {
	      Hooks hooks_;
	      hooks_.push_back (slot);
	      hooks[type].insert(std::make_pair (name, hooks_));
	    }
	    else
	    {
	      Hooks& hooks = (*hooks_i).second; 
	      hooks.push_back (slot);
	    }
	  }

	  void
	  DB::sqlite_exec_simple (const std::string& sql)
	  {
	    //compile the sql statement
	    int		   status;
	    sqlite3_stmt  *pStmt = NULL;
	    const char    *tail;

	    status = sqlite3_prepare (sqlite,
				sql.c_str(),
				sql.length(),
				&pStmt,
				&tail);		 

	    if (status != SQLITE_OK)
	    {
	      g_message ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.c_str ());
	      throw Bmp::DB::SQL_EXEC_ERROR;
	    }

	    for (status = -1; status != SQLITE_DONE; )
	    {
	      status = sqlite3_step (pStmt);

	      if ((status == SQLITE_ERROR) || (status == SQLITE_MISUSE))
	      {
		g_message ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.c_str ());
		break;
	      }
	    }
	    sqlite3_finalize (pStmt);

	  }


	  DB::DB (std::string name, std::string path, bool truncate) : name (name), path (path)
	  {
	      std::string db_filename (Glib::build_filename (path, name) + ".mlib");	    

	      if (Glib::file_test (db_filename, Glib::FILE_TEST_EXISTS) && truncate)
		{
		  unlink (db_filename.c_str());
		}

	      if (sqlite3_open (db_filename.c_str(), &sqlite) != SQLITE_OK)
		{
		  throw IO_ERROR;
		}

	      sqlite3_update_hook (sqlite, DB::sqlite_update_hook, this);
	  };

	  DB::~DB ()
	  {
	      sqlite3_close (sqlite);
	  };

	  void
	  DB::create_attribute_view     (const std::string&   name,
					 const std::string&   view_name,
					 const std::string&   attribute,
					 const ValueMap&      map)	
	  {
		using namespace std;

		mapmap.insert (std::make_pair (name, map));

		stringstream sql;
		sql << "CREATE VIEW " << name << "." << view_name << " AS SELECT DISTINCT " << attribute << " FROM " << name << " ;"; 
		sqlite_exec_simple (sql.str());
	  }

	  void
	  DB::create_view_manual	(const std::string&  name,
					 const std::string&  view_name,
					 const std::string&  select,
					 const ValueMap&     map)
	  {
		using namespace std;

		mapmap.insert (std::make_pair (view_name, map));

		stringstream sql;
		sql << "CREATE VIEW " << name << "." << view_name << " AS " << select << ";";
		sqlite_exec_simple (sql.str());
	  }

	  void
	  DB::add  (const std::string&       name,
		    const Bmp::DB::DataRow&  row)
	  {
		using namespace std;

		if (mapmap.find (name) == mapmap.end ())
		{
		  g_error ("No value-map for table '%s'", name.c_str());
		}

		const ValueMap& map = (*(mapmap.find(name))).second;

		stringstream sql_keys,
			     sql_values;
		bool	     had_value; 

		had_value  =  false;
		sql_keys   << "(";
		sql_values << "("; 

		for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
		{
		  const PairValueType &pair = (*iter);
		  const Bmp::DB::DataRow::const_iterator data_iter = row.find (pair.first);

		  if (data_iter != row. end ())
		  {
		    const Variant &value = (*data_iter).second;

		    if (had_value)
		    {
		      sql_keys    << " , ";
		      sql_values  << " , ";
		    }

		    //add key
		    sql_keys   << pair.first;

		    //add it's value
		    sql_values << "'"; 
		    switch (pair.second)
		    {
			case VALUE_TYPE_STRING:
			{
			  char  *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(value).c_str()); 
			  sql_values << attribute_value_escaped; 
			  g_free (attribute_value_escaped);
			  break;
			}
  
			case VALUE_TYPE_INT:
			{
			  sql_values << value; 
			  break;
			}

			case VALUE_TYPE_BOOL:
			{
			  sql_values << value; 
			  break;
			}

			case VALUE_TYPE_REAL:
			{
			  sql_values << value; 
			  break;
			}

		    }
		    sql_values << "'"; 
		    had_value = true;
		  }
		}
		sql_keys    << ")";
		sql_values  << ")";

		stringstream statement;
		statement << "INSERT INTO '" << name << "' " << sql_keys.str() << " VALUES " << sql_values.str() << ";";
		try {
		  sqlite_exec_simple (statement.str ());
		} catch (Bmp::DB::Exception e) {} //FIXME: Propagate in some way 

	  }

	  void
	  DB::del 	(const std::string&	name,
			 const AttributeList&	attributes)
	  {
		using namespace std;

		AttributeList::const_iterator attributesLast (attributes.end ());
		attributesLast--;

		stringstream sql;
		sql << "DELETE FROM '" << name << "' WHERE ";

		for (AttributeList::const_iterator attributeIter  = attributes.begin  (),
						   attributesEnd  = attributes.end    ();
						   attributeIter != attributesEnd ; ++attributeIter)
		  {
		    const Bmp::DB::Attribute& Attr	= (*attributeIter);
		    const bool exact			= boost::get<0>(Attr);
		    const std::string& attribute_name	= boost::get<1>(Attr);
		    const Variant& attribute_value	= boost::get<2>(Attr);
    
		    if (!exact)
		    {
		      if (attribute_value.which() == VALUE_TYPE_STRING)
		      {
			char *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(attribute_value).c_str()); 
			sql << attribute_name << " LIKE '" << attribute_value_escaped << "%' ";
			g_free (attribute_value_escaped);
		      }
		      else
		      {
			sql << attribute_name << " LIKE '" << attribute_value << "%' ";
		      }
		    } 
		    else
		    {
		      if (attribute_value.which() == VALUE_TYPE_STRING)
		      {
			char *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(attribute_value).c_str()); 
			sql << attribute_name << "='" << attribute_value_escaped << "' ";
			g_free (attribute_value_escaped);
		      }
		      else
		      {
			sql << attribute_name << "='" << attribute_value << "' ";
		      }
		    }

		    if (attributeIter != attributesLast) sql << " AND ";
		  }
		  sql << ";";

                //compile the sql statement
                int	      status;
                sqlite3_stmt  *pStmt;
                const char    *tail;
                status = sqlite3_prepare (sqlite,
                                          sql.str().c_str(),
                                          sql.str().length(),
                                          &pStmt,
                                          &tail);		 
  
                if (status != SQLITE_OK)
                {
                    g_message ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.str().c_str ());
                    throw Bmp::DB::SQL_EXEC_ERROR;
                };
  
  
                for (status = -1; status != SQLITE_DONE; )
                {
                    status = sqlite3_step (pStmt);
  
                    switch (status)
                    {
                        case SQLITE_DONE: break;
                        case SQLITE_ERROR: break;
                    } 
  
                } 
                sqlite3_finalize (pStmt);

	  }

	  void 
	  DB::project     (const std::string&	    name,
			   const std::string&	    project, 
			   DataRowV&		    vector,
			   const AttributeList&	    attributes)
	  {
		using namespace std;

		if (mapmap.find (name) == mapmap.end ())
		{
		  g_error ("No value-map for table '%s'", name.c_str());
		}

		const ValueMap& map = (*(mapmap.find(name))).second;

	        ValueMap::const_iterator valueMapLast (map.end());	
		valueMapLast--;

		stringstream sql;
		if (!attributes.empty ())
		{
		  sql << "SELECT DISTINCT " << project << " FROM '" << name << "' WHERE "; 

		  AttributeList::const_iterator attributesLast (attributes.end ());
		  attributesLast--;
		  for (AttributeList::const_iterator attributeIter  = attributes.begin  (),
						 attributesEnd  = attributes.end    ();
						 attributeIter != attributesEnd ; ++attributeIter)
		  {
		    const Bmp::DB::Attribute& Attr	= (*attributeIter);
		    const bool exact			= boost::get<0>(Attr);
		    const std::string& attribute_name	= boost::get<1>(Attr);
		    const Variant& attribute_value = boost::get<2>(Attr);
    
		    if (!exact)
		    {
		      if (attribute_value.which() == VALUE_TYPE_STRING)
		      {
			char *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(attribute_value).c_str()); 
			sql << attribute_name << " LIKE '" << attribute_value_escaped << "%' ";
			g_free (attribute_value_escaped);
		      }
		      else
		      {
			sql << attribute_name << " LIKE '" << attribute_value << "%' ";
		      }
		    }
		    else
		    {
		      if (attribute_value.which() == VALUE_TYPE_STRING)
		      {
			char *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(attribute_value).c_str()); 
			sql << attribute_name << "='" << attribute_value_escaped << "' ";
			g_free (attribute_value_escaped);
		      }
		      else
		      {
			sql << attribute_name << "='" << attribute_value << "' ";
		      }
		    }

		    if (attributeIter != attributesLast) sql << " AND ";
		  } 
		  sql << ";";
		}
		else
		{
		  sql << "SELECT DISTINCT " << project << " FROM '" << name << "';";
		}

		//compile the sql statement
		sqlite3_stmt  *pStmt;
		const char    *tail;
		int	       status;

		status = sqlite3_prepare (sqlite,
					  sql.str().c_str(),
					  sql.str().length(),
					  &pStmt,
					  &tail);		 

		if (status != SQLITE_OK)
		{
		  g_warning ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.str().c_str ());
		  throw Bmp::DB::SQL_EXEC_ERROR;
		};

		for (status = -1; status != SQLITE_DONE; )
		{
		  status = sqlite3_step (pStmt);
		  switch (status)
		  {
		      case SQLITE_ROW:		    
		      {
			Bmp::DB::DataRow row;
			Bmp::DB::ValueType type = (*(map.find (project))).second;

			switch (type)
			{
			      case VALUE_TYPE_BOOL:
			      {
				int value = sqlite3_column_int (pStmt, 0); 
				row.insert (std::make_pair (project, bool(value)));
				break;
			      }

			      case VALUE_TYPE_INT:
			      {
				int value = sqlite3_column_int (pStmt, 0); 
				row.insert (std::make_pair (project, int(value))); 
				break;
			      }

			      case VALUE_TYPE_REAL:
			      {
				double value = sqlite3_column_double (pStmt, 0); 
				row.insert (std::make_pair (project, double(value)));
				break;
			      }

			      case VALUE_TYPE_STRING:
			      {
				const char* value = (const char*)sqlite3_column_text (pStmt, 0); 
				if (value)
				  row.insert (std::make_pair (project, std::string(value)));
				else
				  row.insert (std::make_pair (project, std::string("")));
				break;
			      }

			      default: break;
			}

			vector.push_back (row);
			break;
		      }

		      case SQLITE_DONE: break;
		      case SQLITE_ERROR: break;
		  } 
		} 

	      sqlite3_finalize (pStmt);
	  }

	  void 
	  DB::set	  (const std::string&	    name,
			   const std::string&	    pkey,
			   const std::string&	    pkey_value,
			   const AttributeListSet&  attributes)
	  {
		using namespace std;

		stringstream sql;
		sql << "UPDATE " << name << " SET ";

		for (AttributeListSet::const_iterator attributeIter  = attributes.begin (),
						      attributesEnd  = attributes.end ();
						      attributeIter != attributesEnd;
						      ++attributeIter)	
		{
		    const Bmp::DB::AttributeSet& Attr	= (*attributeIter);
		    const std::string& attribute_name	= boost::get<0>(Attr);
		    const Variant& attribute_value	= boost::get<1>(Attr);

		    sql << attribute_name;
 
		    if (attribute_value.which() == VALUE_TYPE_STRING)
		    {
			char *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(attribute_value).c_str()); 
			sql << attribute_name << "='" << attribute_value_escaped << "' ";
			g_free (attribute_value_escaped);
		    }
		    else
		    {
			sql << attribute_name << "='" << attribute_value << "' ";
		    }
		}

		sql << " WHERE " << pkey << "='" << pkey_value << "';";
	  }

	  void 
	  DB::get_by_rowid (const std::string&	  name,
			    sqlite_int64	  rowid,
			    DataRow&		  row)
	  {
		using namespace std;

		if (mapmap.find (name) == mapmap.end ())
		{
		  g_error ("No value-map for table '%s'", name.c_str());
		}

		const ValueMap& map = (*(mapmap.find(name))).second;

	        ValueMap::const_iterator valueMapLast (map.end());	
		valueMapLast--;

		stringstream sql;
		sql << "SELECT ";

		for (ValueMap::const_iterator valueMapIter = map.begin	  (),
					      valueMapEnd  = map.end	  ();
					      valueMapIter != valueMapEnd ; ++valueMapIter)
		{
		  const PairValueType &pair = (*valueMapIter);
		  sql << pair.first;
		  if (valueMapIter != valueMapLast) sql << ",";
		}

		sql << " FROM '" << name << "' WHERE rowid='" << rowid << "';";

                //compile the sql statement
                int	       status;
                sqlite3_stmt  *pStmt;
                const char    *tail;
                status = sqlite3_prepare (sqlite,
                                          sql.str().c_str(),
                                          sql.str().length(),
                                          &pStmt,
                                          &tail);		 
  
                if (status != SQLITE_OK)
                {
                    g_message ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.str().c_str ());
                    throw Bmp::DB::SQL_EXEC_ERROR;
                };
  
                for (status = -1; status != SQLITE_DONE; )
                {
                    status = sqlite3_step (pStmt);
  
                    switch (status)
                    {
                        case SQLITE_ROW:		    
                        {
                          unsigned int counter = 0;
                          for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
                          {
                            const PairValueType &pair = (*iter);
                            switch (pair.second)
                            {
                                case VALUE_TYPE_BOOL:
                                {
                                  int value = sqlite3_column_int (pStmt, counter++);	      
                                  row.insert (std::make_pair (pair.first, bool(value))); 
                                  break;
                                }
  
                                case VALUE_TYPE_INT:
                                {
                                  int value = sqlite3_column_int (pStmt, counter++);	      
                                  row.insert (std::make_pair (pair.first, int(value))); 
                                  break;
                                }
  
                                case VALUE_TYPE_STRING:
                                {
				  const char* value = (const char*)sqlite3_column_text (pStmt, counter++);	      

				  if (value)
				  {
				    row.insert (std::make_pair (pair.first, std::string(value)));
				  }
				  else
				  {
				    row.insert (std::make_pair (pair.first, std::string("")));
				  }

				  break;
                                }
  
                                case VALUE_TYPE_REAL:
                                {
                                  double value = sqlite3_column_double (pStmt, counter++);	      
                                  row.insert (std::make_pair (pair.first, double(value))); 
                                  break;
                                }
  
                              default: break;
                            }
                          }
                          return;
                        }
  
                        case SQLITE_DONE: break;
                        case SQLITE_ERROR: break;
                    } 
  
                } 
                sqlite3_finalize (pStmt);
	  }

	  void 
	  DB::get         (const std::string&	  name,
			   const AttributeList&   attributes,
			   DataRowV&		  vector)
	  {
		using namespace std;

		if (mapmap.find (name) == mapmap.end ())
		{
		  g_error ("No value-map for table '%s'", name.c_str());
		}

		const ValueMap& map = (*(mapmap.find(name))).second;

	        ValueMap::const_iterator valueMapLast (map.end());	
		valueMapLast--;

		stringstream sql;
		sql << "SELECT ";

		for (ValueMap::const_iterator valueMapIter = map.begin	  (),
					      valueMapEnd  = map.end	  ();
					      valueMapIter != valueMapEnd ; ++valueMapIter)
		{
		  const PairValueType &pair = (*valueMapIter);
    
		  sql << pair.first;
		  if (valueMapIter != valueMapLast) sql << ",";
		}

		if (!attributes.empty ())
		{
		  sql << " FROM '" << name << "' WHERE ";

		  AttributeList::const_iterator attributesLast (attributes.end ());
		  attributesLast--;
		  for (AttributeList::const_iterator attributeIter  = attributes.begin  (),
						 attributesEnd  = attributes.end    ();
						 attributeIter != attributesEnd ; ++attributeIter)
		  {
		    const Bmp::DB::Attribute& Attr	= (*attributeIter);
		    const bool exact			= boost::get<0>(Attr);
		    const std::string& attribute_name	= boost::get<1>(Attr);
		    const Variant& attribute_value	= boost::get<2>(Attr);
    
		    if (!exact)
		    {
		      if (attribute_value.which() == VALUE_TYPE_STRING)
		      {
			char *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(attribute_value).c_str()); 
			sql << attribute_name << " LIKE '" << attribute_value_escaped << "%' ";
			g_free (attribute_value_escaped);
		      }
		      else
		      {
			sql << attribute_name << " LIKE '" << attribute_value << "%' ";
		      }
		    } 
		    else
		    {
		      if (attribute_value.which() == VALUE_TYPE_STRING)
		      {
			char *attribute_value_escaped = sqlite3_mprintf ("%q", boost::get<std::string>(attribute_value).c_str()); 
			sql << attribute_name << "='" << attribute_value_escaped << "' ";
			g_free (attribute_value_escaped);
		      }
		      else
		      {
			sql << attribute_name << "='" << attribute_value << "' ";
		      }
		    }

		    if (attributeIter != attributesLast) sql << " AND ";
		  }
		  sql << ";";
		}
		else
		{
		  sql << "FROM '" << name << "';";
		}

                //compile the sql statement
                int	     status;
                sqlite3_stmt  *pStmt;
                const char    *tail;
                status = sqlite3_prepare (sqlite,
                                          sql.str().c_str(),
                                          sql.str().length(),
                                          &pStmt,
                                          &tail);		 
  
                if (status != SQLITE_OK)
                {
                    g_message ("SQL Error: '%s', SQL Statement: '%s'", sqlite3_errmsg (sqlite), sql.str().c_str ());
                    throw Bmp::DB::SQL_EXEC_ERROR;
                };
  
  
                for (status = -1; status != SQLITE_DONE; )
                {
                    status = sqlite3_step (pStmt);
  
                    switch (status)
                    {
                        case SQLITE_ROW:		    
                        {
                          Bmp::DB::DataRow row;
                          unsigned int counter = 0;
                          for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
                          {
                            const PairValueType &pair = (*iter);
                            switch (pair.second)
                            {
                                case VALUE_TYPE_BOOL:
                                {
                                  int value = sqlite3_column_int (pStmt, counter++);	      
                                  row.insert (std::make_pair (pair.first, bool(value))); 
                                  break;
                                }
  
                                case VALUE_TYPE_INT:
                                {
                                  int value = sqlite3_column_int (pStmt, counter++);	      
                                  row.insert (std::make_pair (pair.first, int(value))); 
                                  break;
                                }
  
                                case VALUE_TYPE_STRING:
                                {
				  const char* value = (const char*)sqlite3_column_text (pStmt, counter++);	      

				  if (value)
				  {
				    row.insert (std::make_pair (pair.first, std::string(value)));
				  }
				  else
				  {
				    row.insert (std::make_pair (pair.first, std::string("")));
				  }

				  break;
                                }
  
                                case VALUE_TYPE_REAL:
                                {
                                  double value = sqlite3_column_double (pStmt, counter++);	      
                                  row.insert (std::make_pair (pair.first, double(value))); 
                                  break;
                                }
  
                              default: break;
                            }
                          }
                          vector.push_back (row);
                          break;
                        }
  
                        case SQLITE_DONE: break;
                        case SQLITE_ERROR: break;
                    } 
  
                } 
                sqlite3_finalize (pStmt);
	  }

	  void
	  DB::create_table (const std::string&	  name,
			    const std::string&	  pkey,
			    const ValueMap&	  map)
	  {
		using namespace std;

		mapmap.insert (std::make_pair (name, map));

		ValueMap::const_iterator  last;
		stringstream		  sql;

		sql << "CREATE TABLE IF NOT EXISTS '" << name << "' (";
		last = map.end ();
		last--;

		for (ValueMap::const_iterator iter = map.begin (); iter != map.end (); iter++)
		{
		  const PairValueType &pair = (*iter);

		  if (!pkey.compare (pair.first))
		  {
		    sql << "'" << pair.first << "' " << sql_valuenames[pair.second] << " NOT NULL PRIMARY KEY";
		  }
		  else
		  {
		    sql << "'" << pair.first << "' " << sql_valuenames[pair.second] << " "; 
		  }

		  if (iter != last) sql << ",";
		}

		sql << ");";

		try {
		  sqlite_exec_simple (sql.str ());
		} catch (Bmp::DB::Exception e) {} //FIXME: Propagate and check whether the table exists

	  }

	  void
	  DB::drop_table (const std::string& name)
	  {
		using namespace std;

		stringstream sql;
		sql << "DROP TABLE '" << name << "';";

		try {
		  sqlite_exec_simple (sql.str ());
		} catch (Bmp::DB::Exception e) {} //FIXME: Propagate somehow 

	  };


      } // DB

}; // Bmp
