#include <glibmm.h>
#include <glibmm/i18n.h>
#include <gtkmm.h>

#include <config.h>
#include <stdio.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>
#include <libhal-storage.h>
#include <goa/libgoa.h>

#include "hal.hpp"
#include "debug.hpp"

#include "x_system_control.hpp"

namespace
{
  const char* HALBus[] = {
    "<Unknown>",
    "IDE",
    "SCSI",
    "USB",
    "IEEE1394",
    "CCW"
  };

  void
  print_hal_volume (const Bmp::Library::HALVolume& volume)
  {
    g_message ("Volume UDI [%s]", volume.volume_udi.c_str());
    g_message ("\tDevice UDI....: [%s]", volume.device_udi.c_str());
    g_message ("\tMount Path....: [%s]", volume.mount_path.c_str());
    g_message ("\tDevice Serial.: [%s]", volume.device_serial.c_str());
    g_message ("\tVolume Name...: [%s]", volume.volume_name.c_str());
    g_message ("\tDevice File...: [%s]", volume.device_file.c_str());
    g_message ("\tDisc..........: [%s]", volume.is_disc ? "TRUE" : "FALSE");
    g_message ("\tBus...........: [%s]", HALBus[volume.device_bus]); 
  }
}

namespace Bmp
{
  namespace Library
  {
      HAL::HAL () : dbus_connection (0), ctx (0)
      {
	  msg_domain_id = bmp_system_control_message_domain_register (bmp_system_control, "hal", _("BMP: Volume/Device Management"));

	  if (!hal_init())
	  {
	    bmp_system_control_message_dispatch (bmp_system_control,
						 msg_domain_id,
						 GTK_MESSAGE_ERROR,
						 _("Volume/Device Management could not be initialized; you will not"
						   "be able to add items permanently to the library.\n\nPlease check "
						   "your HAL installation and/or contact your distributor."));
	    initialized = false;
	  }
	  else
	  {
	    initialized = true;
	  }
      }

      void
      HAL::scan_devices (const char *udi)
      {
	debug ("hal", "Processing UDI '%s'", udi);  

	char **path_elements = g_strsplit (udi+1, "/", -1);
	if (!g_ascii_strncasecmp (path_elements[4], "volume", strlen ("volume")))
	  {
	    LibHalVolume    *hal_volume;
	    LibHalDrive	    *hal_drive;
	    HALVolume	     volume;

	    volume.volume_udi = udi; 
	    hal_volume = libhal_volume_from_udi (ctx, udi); 

	    if (hal_volume)
	      {
		if (!libhal_volume_is_mounted (hal_volume))
		  {
		    libhal_volume_free (hal_volume);
		    g_strfreev (path_elements);
		    return;
		  }

		volume.mount_path     = libhal_volume_get_mount_point (hal_volume);
		volume.is_disc	      = libhal_volume_is_disc (hal_volume);
		volume.volume_name    = (libhal_volume_get_label (hal_volume)) ? libhal_volume_get_label (hal_volume) : "";
		volume.device_file    = (libhal_volume_get_device_file (hal_volume)) ? libhal_volume_get_device_file (hal_volume) : "";
		volume.device_udi     = libhal_volume_get_storage_device_udi (hal_volume);
		hal_drive = libhal_drive_from_udi (ctx, volume.device_udi.c_str());
		volume.device_serial  = (libhal_drive_get_serial (hal_drive)) ? libhal_drive_get_serial (hal_drive) : "";
		volume.device_bus     = libhal_drive_get_bus (hal_drive);

		libhal_volume_free (hal_volume);
		libhal_drive_free  (hal_drive);
	      }
	    else
	      {
		g_log (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, G_STRLOC ": Unable to retrieve information from HAL for volume UDI '%s'", udi);
	      }
	
	    volumes.insert (std::make_pair (volume.mount_path, volume)); 
#if 0
	    print_hal_volume (volume);
	    g_signal_emit (G_OBJECT(hal), signals[BMP_HAL_SIGNAL_VOLUME_ADDED], 0);
#endif

	  }
	if ((!g_ascii_strncasecmp (path_elements[4], "storage", strlen ("storage"))) || g_strrstr (path_elements[4], "ide"))
	  { 
	    LibHalDrive	*hal_drive = libhal_drive_from_udi (ctx, udi); 
	    if (hal_drive)
	      { 
		// Check for optical device
		if (libhal_drive_get_type (hal_drive) == LIBHAL_DRIVE_TYPE_CDROM)
		{
		    HALDevice device;

		    device.device_udi	= udi;
		    device.device_name	= libhal_drive_get_model (hal_drive);
		    device.device_file  = libhal_drive_get_device_file (hal_drive);

		    optical.push_back (device);
		}
	      }
	    libhal_drive_free  (hal_drive);
	  }
	g_strfreev (path_elements);
      }

      bool
      HAL::hal_init ()
      {
        DBusError     error;
        char	    **devices;
        int	      nr;
	int	      n;

        if (!(ctx = libhal_ctx_new ()))
	{
	  debug ("hal", G_STRLOC ": Failed to initialize HAL!");
	  ctx = 0;
	  return false;
	}

        dbus_error_init (&error);
        if (!hal_mainloop_integration (ctx, &error))
	{
	  debug ("hal", G_STRLOC ": Mainloop integration failed!");
	  dbus_error_free (&error);
	  return false;
	}
#if 0

        libhal_ctx_set_device_new_capability (ctx, hal_device_new_capability);
        libhal_ctx_set_device_lost_capability (ctx, hal_device_lost_capability);
        libhal_ctx_set_device_condition (ctx, hal_device_condition);
#endif

#if 000
        libhal_ctx_set_device_added		(ctx,
						 hal_device_added);

        libhal_ctx_set_device_removed		(ctx,
						 hal_device_removed);

        libhal_ctx_set_device_property_modified	(ctx,
						 hal_property_modified);
#endif //000

        if (!libhal_device_property_watch_all (ctx, &error))
	{
	  debug ("hal", G_STRLOC ": Failed to set up HAL properties watch: %s", error.message);
	  dbus_error_free (&error);
	  libhal_ctx_free (ctx);
	  ctx = 0;
	  return false;
	}

        if (!libhal_ctx_init (ctx, &error))
	{
	  debug ("hal", G_STRLOC ": hal_ctx_init failed: %s", error.message);
	  dbus_error_free (&error);
	  libhal_ctx_free (ctx);
	  ctx = 0;
	  return false;
	}

        if (!(devices = libhal_get_all_devices (ctx, &nr, &error)))
	{
	  debug ("hal", G_STRLOC ": get_all_devices failed: %s", error.message);
	  dbus_error_free (&error);
	  libhal_ctx_shutdown (ctx, NULL);
	  libhal_ctx_free (ctx);
	  ctx = 0;
	  return false;
	}

	for (n = 0; n < nr; n++)
	{
	  scan_devices (devices[n]);
	}

        libhal_free_string_array (devices);
	return true;

      }

      dbus_bool_t
      HAL::hal_mainloop_integration (LibHalContext *ctx,
				     DBusError     *error)
      {
        dbus_connection = dbus_bus_get (DBUS_BUS_SYSTEM, error);

        if (dbus_error_is_set (error))
	{
	  debug ("hal", G_STRLOC ": Error conntecting to DBus");
	  dbus_connection = 0;
	  return false;
	}

        dbus_connection_setup_with_g_main (dbus_connection, NULL);
        libhal_ctx_set_dbus_connection (ctx, dbus_connection);
        return true;
      }

      HAL::~HAL ()
      {
	DBusError error;
	dbus_error_init (&error);

	if (ctx)
	  {
	    libhal_ctx_shutdown (ctx, &error);
	    libhal_ctx_free (ctx);
	  }
	if (dbus_connection)
	  {
	    dbus_connection_close (dbus_connection);
	  }
      }
  }
}

#if 0
static void
hal_device_added (LibHalContext *ctx,
                  const char	*udi)
{
	// BmpHal *hal = BMP_HAL (libhal_ctx_get_user_data (ctx));
}

static void
hal_device_removed (LibHalContext   *ctx,
                    const char	    *udi)
{
	// BmpHal *hal = BMP_HAL (libhal_ctx_get_user_data (ctx));

}

static void
hal_property_modified (LibHalContext	*ctx,
                       const char	*udi,
                       const char	*key,
                       dbus_bool_t       is_removed,
                       dbus_bool_t	 is_added)
{
	BmpHal	  *hal = BMP_HAL (libhal_ctx_get_user_data (ctx));
        // char	  *mounting_udi;
        gboolean   mounted;
        // GSList	  *l, *n;

        if (strcmp (key, "volume.is_mounted") != 0)
                return;

        mounted = libhal_device_get_property_bool (ctx, udi, key, NULL);

        if (mounted)
	  {
	    bmp_hal_add_volume (hal, udi);
	  }
	else
	  {
	    BmpHalVolumeData *v_data;

	    v_data = reinterpret_cast<BmpHalVolumeData*>(g_hash_table_lookup (hal->priv->volumes, udi));

	    if (v_data)
	      {
#if 0
		g_message ("REMOVING:");
		print_bmp_hal_volume_data (v_data);
#endif
		g_hash_table_remove (hal->priv->volumes, udi);
	      }

	    g_signal_emit (G_OBJECT(hal), signals[BMP_HAL_SIGNAL_VOLUME_REMOVED], 0);
	  }
}

static gboolean
v_data_for_each (gpointer _k,
		 gpointer _v,
		 gpointer _d)
{
    BmpHalVolumeData *_v_data;

    _v_data = BMP_HAL_VOLUME_DATA(_v);

    if (!g_ascii_strncasecmp ((gchar*)_d, _v_data->mount_path, strlen (_v_data->mount_path)))
      {
	  return TRUE;
      }

    return FALSE;
}

gboolean
bmp_hal_find_volume_data_for_path (BmpHal	     *hal,
				   const gchar	     *path,
				   BmpHalVolumeData **v_data)
{
    BmpHalVolumeData *_v_data;

    if (!hal->priv->initialized) return FALSE;

    _v_data = reinterpret_cast<BmpHalVolumeData*>(g_hash_table_find (hal->priv->volumes,
		       (GHRFunc)v_data_for_each,
		       (gpointer)path)); 

    if (_v_data)
      {
	  *v_data = _v_data;
	  return TRUE;
      }
    else
      {
	  *v_data = NULL;
	  return FALSE;
      }

    g_assert_not_reached ();
}


gboolean
bmp_hal_volume_is_present (BmpHal	*hal,
			   const gchar	*udi)
{
    return (g_hash_table_lookup (hal->priv->volumes, udi) != NULL) ? TRUE : FALSE;
}

#if 0
typedef gboolean (* DeviceAddedHandler) (const char *udi, const char *capability);
static  gboolean block_device_added (const char *udi, const char *capability __attribute__((__unused__)));

static struct {
        const char *capability;
        DeviceAddedHandler handler;
} devices[] = {
        { "block",          block_device_added   }
};

inline static int
strptrcmp (const void *strptr0, const void *strptr1) {
        return strcmp (*((const char **) strptr0), *((const char **) strptr1));
}



/*
 * cdrom_policy - There has been a media change event on the CD-ROM
 * associated with the given UDI.  Enforce policy.
 */
static void
cdrom_policy (const char *udi)
{
        char *device = NULL;
        char *drive_udi = NULL;
        dbus_bool_t has_audio;
        dbus_bool_t has_data;
        dbus_bool_t is_blank;
        DBusError error;
        int type;

	g_message (G_STRFUNC);

        dbus_error_init (&error);

        has_audio = libhal_device_get_property_bool (hal_ctx, udi, "volume.disc.has_audio", NULL);
        has_data = libhal_device_get_property_bool (hal_ctx, udi, "volume.disc.has_data", NULL);
        is_blank = libhal_device_get_property_bool (hal_ctx, udi, "volume.disc.is_blank", NULL);
        drive_udi = libhal_device_get_property_string (hal_ctx, udi, "info.parent", NULL);
        device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error);

        if (!device) {
                g_message ("cannot get block.device: %s", error.message);
                dbus_error_free (&error);
                goto out;
        }

        if (has_audio && !has_data) {
			g_message ("CD is CDDA and has no data");
        } else if (has_audio && has_data) {
			g_message ("CD is CDDA/Mixed-Mode");
        } else if (has_data) {
			g_message ("CD is CDROM");
        } else if (is_blank) {
			g_message ("CD is blank CD-R");
        }

        /** @todo enforce policy for all the new disc types now supported */

 out:

        libhal_free_string (device);
        libhal_free_string (drive_udi);
}

/*
 * gvm_media_changed - generic media change handler.
 *
 * This is called on a UDI and the media's parent device in response to a media
 * change event.  We have to decipher the storage media type to run the
 * appropriate media-present check.  Then, if there is indeed media in the
 * drive, we enforce the appropriate policy.
 *
 * At the moment, we only handle CD-ROM and DVD drives.
 *
 * Returns TRUE if the device was handled or FALSE otherwise
 */
static gboolean
media_changed (const char *udi, const char *storage_device)
{
        gboolean handled = FALSE;
        char *media_type;
        DBusError error;

	g_message (G_STRFUNC);

        /* Refuse to enforce policy on removable media if drive is locked */
        dbus_error_init (&error);
        if (libhal_device_property_exists (hal_ctx, storage_device, "info.locked", NULL)
            && libhal_device_get_property_bool (hal_ctx, storage_device, "info.locked", NULL)) {
                g_message ("Drive with udi %s is locked through hal; skipping policy\n", storage_device);
                /* we return TRUE here because the device is locked - we can pretend we handled it */
                return TRUE;
        }

        /*
         * Get HAL's interpretation of our media type.  Note that we must check
         * the storage device and not hal UDI
         */
        if (!(media_type = libhal_device_get_property_string (hal_ctx, storage_device, "storage.drive_type", &error))) {
                g_message ("cannot get storage.drive_type: %s", error.message);
                dbus_error_free (&error);

                /* FIXME: should we really return TRUE here? not sure... need a test case */

                return TRUE;
        }

        if (!strcmp (media_type, "cdrom")) {
                cdrom_policy (udi);
                handled = TRUE;
        }

        libhal_free_string (media_type);

        return handled;
}


static gboolean
block_device_added (const char *udi, const char *capability __attribute__((__unused__)))
{
        char *fsusage = NULL, *device = NULL, *storage_device = NULL;
        DBusError error;
        int mountable;

	g_message (G_STRFUNC);

        dbus_error_init (&error);

        /* is hal a mountable volume? */
        if (!(mountable = libhal_device_get_property_bool (hal_ctx, udi, "block.is_volume", NULL))) {
                g_message ("not a mountable volume: %s\n", udi);
                goto out;
        }

        /* if it is a volume, it must have a device node */
        if (!(device = libhal_device_get_property_string (hal_ctx, udi, "block.device", &error))) {
                g_message ("cannot get block.device: %s\n", error.message);
                goto out;
        }

        if (mountable) {
                /* only mount if the block device has a sensible filesystem */
                fsusage = libhal_device_get_property_string (hal_ctx, udi, "volume.fsusage", &error);
                if (!fsusage || strcmp (fsusage, "filesystem") != 0) {
                        g_message ("no sensible filesystem for %s\n", udi);
                        mountable = FALSE;
                }
        }

        /* get the backing storage device */
        if (!(storage_device = libhal_device_get_property_string (hal_ctx, udi, "block.storage_device", &error))) {
                g_message ("cannot get block.storage_device: %s\n", error.message);
                goto out;
        }

        /*
         * Does hal device support removable media?  Note that we
         * check storage_device and not our own UDI
         */
        if (libhal_device_get_property_bool (hal_ctx, storage_device, "storage.removable", NULL)) {
                /* we handle media change events separately */
                g_message ("Changed: %s\n", device);
                if (media_changed (udi, storage_device))
                        goto out;
        }

#if 0
        if (config.automount_drives && mountable) {
                if (!gvm_udi_is_subfs_mount (udi)) {
                        if (gvm_automount_enabled (udi)) {
                                gvm_device_mount (udi, device, NULL);
                        } else {
                                dbg ("storage.automount_enabled_hint set to false on %s, not mounting\n", udi);
                        }
                }
        }
#endif

out:

        if (dbus_error_is_set (&error))
                dbus_error_free (&error);

        libhal_free_string (device);
        libhal_free_string (fsusage);
        libhal_free_string (storage_device);

        return TRUE;
}

/** Invoked when a device in the GDL emits a condition that cannot be
 *  expressed in a property (like when the processor is overheating)
 *
 *  @param  ctx                 LibHal context
 *  @param  udi                 Univerisal Device Id
 *  @param  condition_name      Name of condition
 *  @param  message             D-BUS message with parameters
 */
static void
hal_device_condition (LibHalContext *ctx __attribute__((__unused__)),
                      const char *udi __attribute__((__unused__)),
                      const char *condition_name __attribute__((__unused__)),
                      const char *condition_detail __attribute__((__unused__)))
{
}

static void
hal_device_new_capability (LibHalContext *ctx __attribute__((__unused__)),
                           const char *udi __attribute__((__unused__)),
                           const char *capability __attribute__((__unused__)))
{
}

/** Invoked when device in the Global Device List loses a capability.
 *
 *  @param  ctx                 LibHal context
 *  @param  udi                 Universal Device Id
 *  @param  capability          Name of capability
 */
static void
hal_device_lost_capability (LibHalContext *ctx __attribute__((__unused__)),
                            const char *udi __attribute__((__unused__)),
                            const char *capability __attribute__((__unused__)))
{
	g_message (G_STRFUNC);
}
#endif
#endif
