//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <boost/algorithm/string/trim.hpp>

// Musicbrainz
#include <musicbrainz/musicbrainz.h>

#include <glib/gi18n.h>
#include <gtkmm.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <unistd.h>
#include <sys/stat.h>

#include <iostream>
#include <sstream>

#include <iostream>
#include <boost/shared_ptr.hpp>
#include <boost/format.hpp>

// BMP
#include "uri++.hpp"
#include "database.hpp"
#include "library.hpp"

// TagLib
#include <taglib.h>
#include <fileref.h>
#include <tag.h>
#include <audioproperties.h>
#include <tfile.h>

// GStreamer
#include <gst/gst.h>
#include <gst/gstelement.h>

#include "main.hpp"
#include "paths.hpp"
#include "debug.hpp"

#include "x_library.hpp"

#define TAG_READ_TUPLE(_data) ((TagReadTuple*)(_data))

namespace Bmp
{
  namespace Library
  {
    namespace
    {
      struct TagReadTuple
      {
          DB::DataRow *row;
          bool         eos;
      };

      // We're using GST-style tag identifiers here, except that we can't use dashes as that seems to confuse sqlite
      DatumDefine track_attributes[] = {
          {
              N_("Location"),
              N_("Locations"),
              N_("The location of the track (file, internet stream, etc.)"),
              "location",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Artist"),
              N_("Artists"),
              N_("The artist/peformer of the track"),
              "artist",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Album"),
              N_("Albums"),
              N_("The album of the track"),
              "album",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Title"),
              N_("Titles"),
              N_("The title of the current track"),
              "title",
              DB::VALUE_TYPE_STRING,
              true,
          },

          {
              N_("Track"),
              N_("Tracks"),
              N_("The track number of the track"),
              "tracknumber",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Time"),
              N_("Times"),
              N_("The length of the track"),
              "time",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Genre"),
              N_("Genres"),
              N_("The genre (kind of music) of the track"),
              "genre",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Comment"),
              N_("Comments"),
              N_("Comments added to the track"),
              "comment",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Rating"),
              N_("Ratings"),
              N_("The rating of the track"),
              "rating",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Date"),
              N_("Dates"),
              N_("The date (year) of the track"),
              "date",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("MTIME"),
              N_("MTIMEs"),
              N_("The modification time of the track"),
              "mtime",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Bitrate"),
              N_("Bitrates"),
              N_("The bitrate of the track"),
              "bitrate",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Samplerate"),
              N_("Samplerates"),
              N_("The samplerate of the track"),
              "samplerate",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Play count"),
              N_("Play counts"),
              N_("The number of times this track was played"),
              "count",
              DB::VALUE_TYPE_INT,
              false,
          },

          {
              N_("Volume UDI"),
              N_("Volume UDIs"),
              N_("HAL Volume UDI this track resides on"),
              "volume_udi",
              DB::VALUE_TYPE_STRING,
              false,
          },

          {
              N_("Device UDI"),
              N_("Device UDIs"),
              N_("HAL Device UDI this track resides on"),
              "device_udi",
              DB::VALUE_TYPE_STRING,
              false,
          }
      };

      int
      get_duration_from_sectors (int sectors)
      {
          const int bytes_per_sector   = 2352;
          const int bytes_per_seconds  = (44100 / 8) / 16 / 2;

          return (sectors * bytes_per_sector / bytes_per_seconds)/1000;
      }

      void
      tag_foreach (const GstTagList *list,
                   const gchar      *tag,
                   TagReadTuple     *tuple)
      {
          int count = gst_tag_list_get_tag_size (list, tag);

          for (int i = 0; i < count; i++)
          {
	      const GValue *value = gst_tag_list_get_value_index (list, tag, i);

	      if (G_VALUE_HOLDS (value, G_TYPE_INT))
	      {
		  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_int(value))));
	      }
	      else
	      if (G_VALUE_HOLDS (value, G_TYPE_UINT))
	      {
		  tuple->row->insert (std::make_pair (std::string (tag), int (g_value_get_uint(value))));
	      }
	      else
	      if (G_VALUE_HOLDS (value, G_TYPE_STRING))
	      {
		  std::string v = g_value_get_string (value);
		  boost::trim (v);
		  tuple->row->insert (std::make_pair (std::string (tag), v)); 
	      }
          }
      }

      gboolean
      metadata_bus_handler (GstBus       *bus,
                            GstMessage   *message,
                            TagReadTuple *tuple)
      {
          switch (GST_MESSAGE_TYPE (message))
          {
          case GST_MESSAGE_EOS:
              {
                  tuple->eos = true;
              }
              break;

          case GST_MESSAGE_TAG:
              {
                  GstTagList *tag_list = 0;

                  gst_message_parse_tag (message, &tag_list);
                  if (tag_list)
                  {
                      gst_tag_list_foreach (tag_list, (GstTagForeachFunc) tag_foreach, tuple);
                      gst_tag_list_free (tag_list);
                  }
              }
              break;

          case GST_MESSAGE_ERROR:
              {
                  GError *error = 0;
                  gst_message_parse_error (message, &error, 0);
                  g_print ("%s ERROR: %s\n", G_STRLOC, error->message);
                  g_error_free (error);
              }
              break;

          default:
              break;
          }

          return FALSE;
      }

      void
      metadata_event_loop (TagReadTuple *tuple,
                           GstElement   *element,
                           gboolean      block)
      {
          GstBus *bus = gst_element_get_bus (element);
          g_return_if_fail (bus != 0);

          bool done = false;

          while (!done && !tuple->eos)
          {
              GstMessage *message;

              if (block)
                  message = gst_bus_poll (bus, GST_MESSAGE_ANY, -1);
              else
                  message = gst_bus_pop  (bus);

              if (message == 0)
              {
                  gst_object_unref (bus);
                  return;
              }

              done = metadata_bus_handler (bus, message, tuple);
              gst_message_unref (message);
          }
          gst_object_unref (bus);
      }

    } // anonymous namespace

    void
    get_audiocd_metadata_offline (MusicBrainz &o, DB::DataRowV *vector)
    {
        if (o.Query (std::string (MBQ_GetCDTOC)))
        {
            g_log (BMPX_SYSTEM_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: Unable to get read CD TOC", G_STRLOC);
	    return;
        }

        unsigned int n_tracks = o.DataInt (MBE_TOCGetLastTrack);

        for (unsigned int i = 1; i <= n_tracks; i++)
        {
            DB::DataRow row;

            std::string uri = (boost::format ("cdda:///%d") % i).str ();
            std::string title = (boost::format ("Track %d") % i).str ();

            row.insert (std::make_pair (library->get_metadatum_id(DATUM_LOCATION), uri));
            row.insert (std::make_pair (library->get_metadatum_id(DATUM_TITLE), title));
            row.insert (std::make_pair (library->get_metadatum_id(DATUM_TRACK), int(i)));
            row.insert (std::make_pair (library->get_metadatum_id(DATUM_TIME),  int (get_duration_from_sectors(o.DataInt(MBE_TOCGetTrackNumSectors, i)))));

            vector->push_back (row);
        }
    }

    void
    get_audiocd_metadata (int &status, DB::DataRowV &vector)
    {
        // Create the musicbrainz object, which will be needed for subsequent calls
        MusicBrainz o;
        o.UseUTF8 (true);
	o.SetDevice (mcs->key_get<std::string>("audio", "cdrom-device"));

	bool isMultipleArtist = false;
	int trackNum = 0;

        int ret;
        ret = o.Query (MBQ_GetCDTOC);
        if (!ret)
        {
            string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
	    return;
        }

        std::string cdindex = o.Data(std::string (MBE_TOCGetCDIndexId));
        if (cdindex.empty())
        {
            std::string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
	    return;
        }

        ret = o.Query(std::string (MBQ_GetCDInfo));
        if (!ret)
        {
            std::string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
	    return;
        }

        // Select the first album
        std::list<int> argList;
        argList.push_back (1);
        if (!o.Select (std::string (MBS_SelectAlbum), &argList))
        {
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
	    return;
        }

        std::string album_id_url = o.Data(std::string (MBE_AlbumGetAlbumId));
        if (album_id_url.empty ()) return;

        std::string album_id;
        o.GetIDFromURL (album_id_url, album_id);

        // Extract the album name
        std::string album_name = o.Data(std::string (MBE_AlbumGetAlbumName));

        // Extract the number of tracks
        int n_tracks = o.DataInt(std::string (MBE_AlbumGetNumTracks));

        if (n_tracks < 1)
        {
            status = Bmp::Library::Library::MB_OFFLINE;
            get_audiocd_metadata_offline (o, &vector);
	    return;
        }

        // Check to see if there is more than one artist for this album
        std::string artist_check;
        for (int i = 1; i <= n_tracks; i++)
        {
            std::string artist;
            artist = o.Data (std::string (MBE_AlbumGetArtistId));
            if (artist.empty ())
                break;

            if (i == 1)
            {
                artist_check = artist;
                continue;
            }

            if (artist.compare (artist_check))
            {
                isMultipleArtist = true;
                break;
            }
        }

        std::string artist_name, artist_id;

        if (!isMultipleArtist)
        {
            // Extract the artist name from the album
            artist_name = o.Data (std::string (MBE_AlbumGetArtistName), 1);

            // Extract the artist id from the album
            artist_id   = o.Data (std::string (MBE_AlbumGetArtistId), 1);
        }

        for(int i = 1; i <= n_tracks; i++)
        {
            DB::DataRow row;
            std::stringstream uri;

            uri << "cdda://" << cdindex << "/" << i;

            row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id, uri.str()));
            row.insert (std::make_pair (track_attributes[DATUM_ALBUM].id, std::string (album_name)));
            row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id, int(1411)));
            row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int(44100)));

            // Extract the track name from the album.
            std::string title = o.Data (std::string (MBE_AlbumGetTrackName), i);
            if (!title.empty ())
            {
                row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, DB::Variant (std::string (title))));
            }

            // Extract the album id from the track. Just use the
            // first album that this track appears on
            std::string track_id = o.Data (std::string (MBE_AlbumGetTrackId), i);
            if (!track_id.empty ())
            {
                // Extract the track number
                trackNum = o.GetOrdinalFromList (std::string (MBE_AlbumGetTrackList), track_id);
                if (trackNum > 0 && trackNum < 100)
                {
                    row.insert (std::make_pair (track_attributes[DATUM_TRACK].id,
                                                 DB::Variant (int(trackNum))));
                }
            }

            // If its a multple artist album, print out the artist for each track
            if (isMultipleArtist)
            {
                // Extract the artist name from this track
                string individual_artist;
                individual_artist = o.Data(std::string (MBE_AlbumGetArtistName), i);

                if (!individual_artist.empty ())
                {
                    row.insert (std::make_pair (track_attributes[DATUM_ARTIST].id,
                                                 DB::Variant (std::string (individual_artist))));
                }
            }
            else
            {
                row.insert (std::make_pair (track_attributes[DATUM_ARTIST].id,
                                             DB::Variant (std::string (artist_name))));
            }

            vector.push_back (row);
        }

        ret = o.Query(MBQ_GetCDTOC);
        if (!ret)
        {
            std::string error;
            o.GetQueryError (error);
            debug ("library", "%s: Query failed: %s", G_STRLOC, error.c_str());
            status = Bmp::Library::Library::MB_FAILURE;
	    return;
        }

        for(int i = 1; i <= n_tracks; i++)
        {
            DB::DataRow& row = vector[i-1];
            int duration = get_duration_from_sectors(o.DataInt(MBE_TOCGetTrackNumSectors, i+1));
            row.insert (std::make_pair (track_attributes[DATUM_TIME].id, duration)); 
	} 

        status = Bmp::Library::Library::MB_OK;
    }

    bool 
    Library::metadata_set_taglib (const UpdateTrack &track)
    {
	Bmp::URI u (track.location.get());
	u.unescape ();
        std::string filename = u.path;

        if (filename.empty()) return false;

        TagLib::File *file = 0;
	file = TagLib::FileRef::create(filename.c_str());
  
	bool result = false;
	if ( file && file->isValid() && file->tag() && !file->readOnly())
	{
	  if (track.title) file->tag()->setTitle(track.title.get());
	  if (track.artist) file->tag()->setArtist(track.artist.get());
	  if (track.album) file->tag()->setAlbum(track.album.get());
	  if (track.genre) file->tag()->setGenre(track.genre.get());
	  if (track.comment) file->tag()->setComment(track.comment.get());
	  if (track.tracknumber) file->tag()->setTrack(track.tracknumber.get());
	  if (track.date) file->tag()->setYear(track.date.get());
	  file->save();
	  result = true;
	}

	if (file) delete file;	
	return result;
    }

    void 
    Library::metadata_get_taglib (const	  std::string &uri,
				  DB::DataRow	      &row)
    {
	Bmp::URI u (uri);
	u.unescape ();
        std::string filename = u.path;

        if (filename.empty()) return;

        TagLib::File *track = TagLib::FileRef::create(filename.c_str());
        if ((!track) || (!track->tag()) || (track->tag()->isEmpty())) return;

        struct stat file_stat;
        stat (filename.c_str(), &file_stat);

	using namespace boost;
	std::string value;

        row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id,   std::string(uri)));
        row.insert (std::make_pair (track_attributes[DATUM_TRACK].id,      int (track->tag()->track())));
        row.insert (std::make_pair (track_attributes[DATUM_DATE].id,       int (track->tag()->year())));

        value = track->tag()->artist().to8Bit (true);
	trim (value);
        row.insert (std::make_pair (track_attributes[DATUM_ARTIST].id, value)); 

        value = track->tag()->album().to8Bit (true);
	trim (value);
        row.insert (std::make_pair (track_attributes[DATUM_ALBUM].id, value)); 

        value = track->tag()->genre().to8Bit (true);
	trim (value);
        row.insert (std::make_pair (track_attributes[DATUM_GENRE].id, value));

        value = track->tag()->comment().to8Bit (true);
	trim (value);
        row.insert (std::make_pair (track_attributes[DATUM_COMMENT].id, value)); 

        row.insert (std::make_pair (track_attributes[DATUM_RATING].id,     int (0)));
        row.insert (std::make_pair (track_attributes[DATUM_COUNT].id,      int (0)));
        row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id,    int (track->audioProperties()->bitrate())));
        row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int (track->audioProperties()->sampleRate())));
        row.insert (std::make_pair (track_attributes[DATUM_TIME].id,       int (track->audioProperties()->length())));

        if  ((!track->tag()->title().toCString(true)) || (!std::strlen(track->tag()->title().toCString(true))))
            row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, std::string (Glib::path_get_basename(u.path))));
        else
	{
            value = track->tag()->title().to8Bit(true);
	    trim (value);
            row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, value)); 
	}

        delete track;
    }

    void 
    Library::metadata_get_gst  (const std::string &uri,
				DB::DataRow	  &row)
    {
        boost::shared_ptr<TagReadTuple> ttuple;

        ttuple = boost::shared_ptr<TagReadTuple> (new TagReadTuple);
        ttuple->eos = false;
	ttuple->row = &row;

        ttuple->row->insert (std::make_pair (get_metadatum_id (DATUM_TITLE),    std::string (uri)));
        ttuple->row->insert (std::make_pair (get_metadatum_id (DATUM_LOCATION), std::string (uri)));

        GstElement *element   = gst_element_factory_make ("playbin",  "playbin0");
        GstElement *fakesink  = gst_element_factory_make ("fakesink", "fakesink0");
        GstElement *videosink = gst_element_factory_make ("fakesink", "fakesink1");

	Bmp::URI u (uri);
	u.unescape ();

        std::string uri_str = std::string (u); 
        g_object_set (G_OBJECT (element), "uri", uri_str.c_str(), "audio-sink", fakesink, "video-sink", videosink, NULL);

        GstStateChangeReturn state_ret = gst_element_set_state (element, GST_STATE_PAUSED);
        gint                 change_timeout = 5;

        while (state_ret == GST_STATE_CHANGE_ASYNC && change_timeout > 0)
        {
            GstState state;
            state_ret = gst_element_get_state (GST_ELEMENT (element), &state, 0, 1 * GST_SECOND);
            change_timeout--;
        }

        if (state_ret != GST_STATE_CHANGE_FAILURE && state_ret != GST_STATE_CHANGE_ASYNC)
        {
            /* Post application specific message so we'll know when to stop
             * the message loop */
            GstBus *bus;
            bus = gst_element_get_bus (GST_ELEMENT (element));
            if (bus)
            {
                gst_bus_post (bus, gst_message_new_application (GST_OBJECT (element), 0));
                gst_object_unref (bus);
            }

            /* Poll the bus for messages */
            metadata_event_loop (ttuple.get(), GST_ELEMENT (element), FALSE);
        }

        GstFormat  format = GST_FORMAT_TIME;
        GstQuery  *query  = gst_query_new_duration (format);

        if (gst_element_query (GST_ELEMENT(element), query))
        {
            gst_element_get_state (GST_ELEMENT(element), 0, 0, 100 * GST_MSECOND);

            gint64 length_in_nanoseconds;
            gst_query_parse_duration (query, &format, &length_in_nanoseconds);

            int length = length_in_nanoseconds / GST_SECOND;
            ttuple->row->insert (std::make_pair (get_metadatum_id (DATUM_TIME), DB::Variant (length)));
        }
        gst_query_unref (query);
        gst_element_set_state (GST_ELEMENT (element), GST_STATE_NULL);
        gst_object_unref (GST_ELEMENT (element));
    }

    std::string
    Library::get_titlestring (const std::string& uri)
    {
      GString	  *string_out   = 0;
      GString	  *string_field = 0;
      char	  *field;
      const char  *format;
      int	   n;
      bool	   hit = false;

      Bmp::DB::DataRow row;

      try { get_metadata (uri, row); } catch (...) { return std::string(); }

      if (row.empty()) return std::string();

      format = mcs->key_get<std::string>( "bmp", "titlestring-format") . c_str ();
      if (!format) return std::string();

      string_out = g_string_new ("");

      while (*format) {

	/* Begin sequence */
	if ((*format == '%') && (*(format+1) != 0) && (*(format+1) == '{'))
	  {
	    string_field = g_string_new ("");
	    format += 2;
	    while (*format && (*format != '}') && (*format != ' '))
	      {
		g_string_append_c (string_field, *format);
		format++;
	      }

	    field = g_string_free (string_field, FALSE);
	    if (!*format)
	      {
	        free (field);
		char *str = g_string_free (string_out, FALSE);
		std::string stdstr = str;
		free (str);
		return stdstr;
	      }

	    /* Append metadata field */
	    for (n = 0; n < Bmp::Library::DATUM_LAST; ++n)
	      {
		if (!g_ascii_strcasecmp (field, get_metadatum_id (Bmp::Library::Datum (n)).c_str())) 
                  {
		    Bmp::DB::Variant v = row[std::string(field)];
		    char *value = 0;

		    if (Bmp::DB::ValueType(v.which()) == Bmp::DB::VALUE_TYPE_INT)
		    {
			value = g_strdup_printf ("%d", boost::get<int>(v)); 
		    }
		    else if (Bmp::DB::ValueType(v.which()) == Bmp::DB::VALUE_TYPE_STRING)
		    {
			value = strdup (boost::get<std::string>(v).c_str());
		    }

		    if (value)
		    {
		      g_strstrip (value);
		      if (strlen(value) && strcmp(value, "0") != 0)
		      {
			  hit = true;
			  g_string_append (string_out, value);
		      }
		      free (value);
		    }
		    break;
		  }
	      }
	    free (field);
	  }
	else
	  {
	    g_string_append_c (string_out, *format);
	  }
	format++;
    }

    if (!hit)
      {
	g_string_free (string_out, TRUE);
	return Glib::path_get_basename (uri);
      }
    else
      {
	char *str = g_string_free (string_out, FALSE);
	std::string stdstr = str;
	free (str);
	return stdstr;
      }
    }

    void 
    Library::cache_metadata (const std::string	    &uri,
			     const std::string	    &root_path)
    {
  
	DB::DataRow row;
	Bmp::URI u (uri);
	u.unescape ();

	debug ("library", "caching URI: '%s'", uri.c_str());

	try { metadata_get_taglib (std::string(u), row); } catch (Bmp::Library::Exception cxe) { metadata_get_gst (std::string(u), row); }

	if (!row.empty ())
	{
	    Track track (row);
	    if (track.artist && track.album && track.title)
	    {
	      db->add ("main", row);
	    }
	}
    }

    void
    Library::del (const DB::AttributeList &attrs)
    {
	signal_processing_start_.emit ();
	db->del("main", attrs);
	signal_processing_end_.emit ();
    }

    void
    Library::query (const DB::AttributeList &attrs,
		    DB::DataRowV	    &vector)
    {
	db->get ("main", attrs, vector);
    }

    bool
    Library::is_compilation (const std::string& album)
    {
	return (compilations.find (album) != compilations.end());
    }

    /** The caller must be aware that each row in this vector contains ONLY the requrested column, re-marshaling/transforming
      * the result into something more suitable, i.e. an std::vector<> of std::string would be just too much overhead
      */
    void
    Library::project (const std::string&	project,
		      DB::DataRowV		&vector,
		      const DB::AttributeList	&attrs)
    {
	return db->project ("main", project, vector, attrs);
    }

    GHashTable*
    Library::get_hashtable (const DB::DataRow& row)
    {
      GHashTable *table;
      GValue	 *value;

      table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

      for (unsigned int n = 0; n < G_N_ELEMENTS (track_attributes); n++)
	{
	  DB::DataRow::const_iterator i = row.find (track_attributes[n].id);
	  if (i == row.end())
	  {
	      g_warning (G_STRLOC ": Attribute '%s' not found in DataRow", track_attributes[n].id);
	      continue;
	  }

	  value = g_new0 (GValue,1);
	  switch (Bmp::DB::ValueType(i->second.which()))
	  {
	    case Bmp::DB::VALUE_TYPE_STRING:
		    g_value_init (value, G_TYPE_STRING);
		    g_value_set_string (value, (boost::get<std::string>(i->second)).c_str());
		    break;

	    case Bmp::DB::VALUE_TYPE_INT:
		    g_value_init (value, G_TYPE_INT);
		    g_value_set_int (value, (boost::get<int>(i->second)));
		    break;

	    case Bmp::DB::VALUE_TYPE_BOOL:
		    g_value_init (value, G_TYPE_BOOLEAN);
		    g_value_set_boolean (value, (boost::get<bool>(i->second)));
		    break;

	    case Bmp::DB::VALUE_TYPE_REAL:
		    g_value_init (value, G_TYPE_DOUBLE);
		    g_value_set_double (value, (boost::get<double>(i->second)));
		    break;
	  }
	  g_hash_table_insert (table, g_strdup(track_attributes[n].id), value);
	}
      return table;
    }

    void
    Library::get_metadata (const std::string &uri,
			   DB::DataRow	     &row)
    {
	Bmp::URI u (uri);
	u.unescape ();

        DB::AttributeList attrs;
        DB::DataRowV vector;
        attrs.push_back (DB::Attribute (true, track_attributes[DATUM_LOCATION].id, std::string(u)));
	query (attrs, vector);

	if (vector.empty())
	{
	  try { metadata_get_taglib (std::string(u), row); } catch (Bmp::Library::Exception& cxe) { metadata_get_gst (std::string(u), row); }

	  if (row.empty ())
	  {
	    row.insert (std::make_pair (track_attributes[DATUM_LOCATION].id, uri)); 
	    row.insert (std::make_pair (track_attributes[DATUM_SAMPLERATE].id, int(0))); 
	    row.insert (std::make_pair (track_attributes[DATUM_BITRATE].id, int(0)));
	    if (u.get_protocol() == Bmp::URI::PROTOCOL_HTTP)
	    {
	      row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, uri));
	    }
	    else
	    {
	      row.insert (std::make_pair (track_attributes[DATUM_TITLE].id, Glib::path_get_basename (std::string(u.path))));
	    }
	  }
	}
	else
	{
	  row = vector[0]; 
	}	
      
    }

    const std::string
    Library::get_metadatum_id (Bmp::Library::Datum metadatum) const
    {
        return std::string (track_attributes[metadatum].id);
    }

    const DatumDefine 
    Library::get_metadatum_define (Bmp::Library::Datum metadatum) const
    {
        return track_attributes[metadatum];
    }


    bool
    Library::cache_audiocd (const std::string& cdindex_id)
    {
        int status;
        DB::DataRowV rows;
        get_audiocd_metadata (status, rows);
        if (status != Bmp::Library::Library::MB_OK) return false;

	DB::AttributeList attrs;
	DB::DataRowV vector;

	std::stringstream value;
	value << "cdda://" << cdindex_id << "/1";
        attrs.push_back (DB::Attribute (true, track_attributes[DATUM_LOCATION].id, value.str ()));
  
	query (attrs, vector);
	if (vector.empty())
	{
	  for (DB::DataRowV::const_iterator iter = rows.begin (); iter != rows.end (); ++iter)
	  {
	    db->add ("main", (*iter));
	  }
	}
        return true;
    }

    Library::Library ()
    {
	using namespace DB;
	using namespace std;

        for (unsigned int n = 0; n < G_N_ELEMENTS(track_attributes); n++)
        {
            map.insert (std::make_pair(std::string (track_attributes[n].id), track_attributes[n].type));
        }

        db = new ::Bmp::DB::DB ("library", BMP_PATH_DATA_DIR, false);
        db->create_table  ("main",				      //table name
			   string (track_attributes[DATUM_LOCATION].id), //pkey identifier/column name
			   map);				      //mapping for the query (must be the same mapping used to create the table)

	stringstream sql;
	sql << "SELECT DISTINCT " << track_attributes[DATUM_ALBUM].id 
	    << ", " 
	    << "count(DISTINCT " << track_attributes[DATUM_ARTIST].id << ") AS nr FROM main GROUP BY "
	    << track_attributes[DATUM_ALBUM].id << " HAVING nr > 1"; 

	map_compilations.insert (make_pair (string (track_attributes[DATUM_ALBUM].id), track_attributes[DATUM_ALBUM].type));
	map_compilations.insert (make_pair (string ("nr"), VALUE_TYPE_INT));

	try {
	  db->create_view_manual ("main", "compilations", sql.str(), map_compilations);
	} catch (...) {}

	update_compilations ();
    }

    void
    Library::update_compilations ()
    {
	using namespace DB;
	using namespace std;

	DataRowV rows;
	db->project ("compilations", track_attributes[DATUM_ALBUM].id, rows);
	compilations.clear ();
	for (DataRowV::iterator i = rows.begin (), end = rows.end (); i != end; ++i) 
	{
	    compilations.insert (boost::get<std::string>((*i)[track_attributes[DATUM_ALBUM].id]));
	}
    }

    void
    Library::modify_tracks (const UpdateTracks& tracks)
    {
	for (UpdateTracks::const_iterator	i_Tracks      = tracks.begin (),
						i_TracksEnd   = tracks.end ();
						i_Tracks != i_TracksEnd; ++i_Tracks)
	{
	    using namespace Bmp::DB;

	    UpdateTrack const& track = (*i_Tracks);
	    if (metadata_set_taglib (track))
	    {
		AttributeListSet attrs;

		if (track.title) attrs.push_back (AttributeSet (track_attributes[DATUM_TITLE].id, std::string(track.title.get()))); 
	        if (track.artist) attrs.push_back (AttributeSet (track_attributes[DATUM_ARTIST].id, std::string(track.artist.get()))); 
		if (track.album) attrs.push_back (AttributeSet (track_attributes[DATUM_ALBUM].id, std::string(track.album.get()))); 
		if (track.genre) attrs.push_back (AttributeSet (track_attributes[DATUM_GENRE].id, std::string(track.genre.get()))); 
		if (track.comment) attrs.push_back (AttributeSet (track_attributes[DATUM_COMMENT].id, std::string(track.comment.get()))); 
		if (track.tracknumber) attrs.push_back (AttributeSet (track_attributes[DATUM_TRACK].id, int(track.tracknumber.get()))); 
		if (track.date) attrs.push_back (AttributeSet (track_attributes[DATUM_DATE].id, int(track.date.get()))); 
		if (track.rating) attrs.push_back (AttributeSet (track_attributes[DATUM_RATING].id, int(track.rating.get()))); 
		/* if (track.mtime) attrs.push_back (AttributeSet (track_attributes[DATUM_MTIME].id, int(track.mtime.get()))); */
		db->set ("main", track_attributes[DATUM_LOCATION].id, track.location.get (), attrs);
	    }
	}	
    }

    Library::~Library ()
    {
        delete db;
    }

    Library::SignalProcessing&
    Library::signal_processing_start ()
    { 
      return signal_processing_start_;
    }

    Library::SignalProcessing&
    Library::signal_processing_end ()
    { 
      return signal_processing_end_;
    }

    //class Bmp::Library::Track
    Track::~Track ()
    {}

    Track::Track (const Bmp::DB::DataRow& row)
    {
	using namespace Bmp::DB;
	DataRow::const_iterator i, end = row.end ();

	i = row.find (track_attributes[DATUM_LOCATION].id);
	if (i != end) this->location = boost::get<std::string> (i->second);

	i = row.find (track_attributes[DATUM_ARTIST].id);
	if (i != end) this->artist = boost::get<std::string> (i->second);

	i = row.find (track_attributes[DATUM_ALBUM].id);
	if (i != end) this->album = boost::get<std::string> (i->second);

	i = row.find (track_attributes[DATUM_TITLE].id);
	if (i != end) this->title = boost::get<std::string> (i->second);

	i = row.find (track_attributes[DATUM_GENRE].id);
	if (i != end) this->genre = boost::get<std::string> (i->second);

	i = row.find (track_attributes[DATUM_COMMENT].id);
	if (i != end) this->comment = boost::get<std::string> (i->second);

	i = row.find (track_attributes[DATUM_TRACK].id);
	if (i != end) this->tracknumber = boost::get<int> (i->second);

	i = row.find (track_attributes[DATUM_TIME].id);
	if (i != end) this->duration = boost::get<int> (i->second);

	i = row.find (track_attributes[DATUM_RATING].id);
	if (i != end) this->rating = boost::get<int> (i->second);

	i = row.find (track_attributes[DATUM_DATE].id);
	if (i != end) this->date = boost::get<int> (i->second);

	i = row.find (track_attributes[DATUM_MTIME].id);
	if (i != end) this->mtime = boost::get<int> (i->second);

	i = row.find (track_attributes[DATUM_BITRATE].id);
	if (i != end) this->bitrate = boost::get<int> (i->second);

	i = row.find (track_attributes[DATUM_SAMPLERATE].id);
	if (i != end) this->samplerate = boost::get<int> (i->second);

	i = row.find (track_attributes[DATUM_COUNT].id);
	if (i != end) this->samplerate = boost::get<int> (i->second);
    }

  } // Library namespace

} // Bmp namespace
