#ifndef BMP_LIBRARY_CLASS_HPP
#define BMP_LIBRARY_CLASS_HPP

#include <gtkmm.h>
#include <glibmm/i18n.h>

#include <boost/optional.hpp>
#include <set>

#include "database.hpp"
#include "util.hpp"

namespace Bmp
{
  namespace Library
  {
    enum Exception
    {
      NO_METADATA,
      UNABLE_TO_HANDLE,
      INVALID_PATH,
    }; 

    enum Datum
    {
        DATUM_LOCATION,
        DATUM_ARTIST,
        DATUM_ALBUM,
        DATUM_TITLE,
        DATUM_TRACK,
        DATUM_TIME,
        DATUM_GENRE,
        DATUM_COMMENT,
        DATUM_RATING,
        DATUM_DATE,
        DATUM_MTIME,
        DATUM_BITRATE,
        DATUM_SAMPLERATE,
        DATUM_COUNT,
        DATUM_HAL_VOLUME_UDI,
        DATUM_HAL_DEVICE_UDI,

	DATUM_LAST
    };

    struct DatumDefine
    {
          const char    *title;
          const char    *title_plural;
          const char    *description;
          const char    *id;
          DB::ValueType  type;
          bool           filter;
    };

    struct Track
    {
	  boost::optional <std::string> location;
	  boost::optional <std::string> artist;
	  boost::optional <std::string>	album;
	  boost::optional <std::string>	title;
	  boost::optional <std::string>	genre;
	  boost::optional <std::string>	comment;
	  boost::optional <std::string>	volume_udi;
	  boost::optional <std::string>	device_udi;

	  boost::optional <int>		tracknumber;
	  boost::optional <int>		duration;
	  boost::optional <int>		rating;
	  boost::optional <int>	  	date;
	  boost::optional <int>		mtime;
	  boost::optional <int>		bitrate;
	  boost::optional <int>		samplerate;
	  boost::optional <int>		count;

	  Track (const Bmp::DB::DataRow& row);
	  Track () {};
	  ~Track ();
    };

    struct UpdateTrack
    {
	  boost::optional <std::string> location;
	  boost::optional <std::string> artist;
	  boost::optional <std::string>	album;
	  boost::optional <std::string>	title;
	  boost::optional <std::string>	genre;
	  boost::optional <std::string>	comment;

	  boost::optional <int>		tracknumber;
	  boost::optional <int>	  	date;
	  boost::optional <int>		rating;
	  boost::optional <int>		mtime;
    };

    typedef std::vector < UpdateTrack > UpdateTracks;

    class Library
    {
      public:
        typedef sigc::signal<void> SignalProcessing;
        SignalProcessing& signal_processing_start ();
        SignalProcessing& signal_processing_end   ();

        enum
        {
            MB_FAILURE,
            MB_OK,
            MB_OFFLINE
        };

	typedef std::set < std::string > Compilations;

        Library ();
        ~Library ();

	void
	query	  (const DB::AttributeList &attrs,
		   DB::DataRowV &vector);

	void
	project	  (const std::string& project,
		   DB::DataRowV &vector,
		   const DB::AttributeList &attrs = DB::AttributeList ());

	bool
	is_compilation (const std::string& album);

	void
	update_compilations ();
 
	void
	del (const DB::AttributeList &attrs);

	void	
        get_metadata (const std::string &pkey,
		      DB::DataRow &row);

	GHashTable*
	get_hashtable (const DB::DataRow& row);

	void
	cache_metadata (const std::string &uri,
			const std::string &path);

        bool
        cache_audiocd (const std::string &cdindex_id);

        const std::string
        get_metadatum_id (Datum metadatum) const;

	const DatumDefine 
	get_metadatum_define (Datum metadatum) const;

	std::string
	get_titlestring (const std::string& uri);

	void
	modify_tracks (const UpdateTracks& tracks);
 
    private:

	SignalProcessing signal_processing_start_;
	SignalProcessing signal_processing_end_;

        DB::ValueMap  map;
        DB::DB       *db;

        DB::ValueMap  map_compilations;
	Compilations  compilations;

	bool 
	metadata_set_taglib (const UpdateTrack &track);
	
	void 
        metadata_get_taglib (const std::string &uri,
			     DB::DataRow       &row);

	void
        metadata_get_gst    (const std::string &uri,
			     DB::DataRow       &row);

    }; // class Library

  } // Library namespace

} // Bmp namespace

#endif // BMP_LIBRARY_CLASS_HPP
