#include <gtkmm.h>
#include <glibmm.h>

#include <iostream>
#include <sstream>

#include <musicbrainz/musicbrainz.h>
#include <boost/format.hpp>
#include <boost/optional.hpp>

#include <chroma/chroma-list.h>
#include <chroma/chroma-list-selection.h>

#include <widgets/bmp_tray_icon.h>
#include <widgets/bmp_status_icon.h>

#include <goa/libgoa.h>

#include "paths.hpp"
#include "main.hpp"

#include "x_library.hpp"
#include "x_play.hpp"
#include "x_playlist.hpp"
#include "x_system_control.hpp"

#include "filebrowser.hpp"
#include "trackinfo.hpp"
#include "preferences.hpp"
#include "streams.hpp"
#include "library_ui.hpp"
#include "about.hpp"
#include "export.hpp"
#include "network.hpp"

#include "ui_util.hpp"

#include "winamp2_skin.hpp"
#include "winamp2_main.hpp"
#include "winamp2_playlist.hpp"

#include "shell.hpp"
#include "action_defines.hpp"

#include "x_resource_manager.hpp"

namespace Bmp
{
  Glib::RefPtr<Gtk::UIManager>	   ui_manager;
  Glib::RefPtr<Gtk::ActionGroup>   actions;
  Glib::RefPtr<Gtk::ActionGroup>   actions_toggle; 

  BmpWinamp2Skin	  *winamp2_skin;
  BmpWindowMain		  *w_main;
  BmpWindowPlaylist	  *w_playlist;
  Mcs::Bind		  *mcs_bind;
}

namespace
{
  bool running = false;

  // Data
  boost::optional<int> x_main;
  boost::optional<int> y_main;
  boost::optional<int> x_pl;
  boost::optional<int> y_pl;
  boost::optional<int> w_pl;
  boost::optional<int> h_pl;

  // Objects  
  Bmp::FileBrowser	  *browser;
  Bmp::TrackInfo	  *trackinfo;
  Bmp::Preferences	  *preferences;
  Bmp::Streams::Dialog	  *streams;
  Bmp::LibraryUI	  *library_ui;
  Bmp::AboutDialog	  *about;
  Bmp::ExportDialog	  *export_dialog;

  BmpStatusIcon		  *status_icon;
  GdkPixbuf		  *status_icon_image_default;
  GdkPixbuf		  *status_icon_image_paused;
  GdkPixbuf		  *status_icon_image_playing;

  /*------------------------------------------------------------------------------------------------*/
  // Tray Icon Callbacks
  void
  status_icon_click (BmpStatusIcon *icon, gpointer data)
  {
    Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					     (Bmp::actions_toggle->get_action (BMP_TOGGLE_ACTION_SHOW_UI));
    bool active = action->get_active(); 
    action->set_active (!active);
  }

  void
  status_icon_scroll_up (BmpStatusIcon *icon, gpointer data)
  {
    int volume;
    volume = g_object_get_int (bmp_system_control, "volume");
    volume += 4;
    volume = (volume > 100) ? 100 : volume;
    bmp_system_control_volume_set (bmp_system_control, volume, NULL);
  }

  void
  status_icon_scroll_down (BmpStatusIcon *icon, gpointer data)
  {
    int volume;
    volume = g_object_get_int (bmp_system_control, "volume");
    volume -= 4;
    volume = (volume < 0) ? 0 : volume;
    bmp_system_control_volume_set (bmp_system_control, volume, NULL);
  }

  void
  status_icon_popup_menu (BmpStatusIcon *icon, guint arg1, guint arg2, gpointer data)
  {
    GtkWidget *menu;
    menu = ui_manager_get_popup (Bmp::ui_manager->gobj(), "/popup-tray/menu-tray");
    gtk_widget_realize (menu);
    gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, arg1, arg2);
  }

  /*------------------------------------------------------------------------------------------------*/
  // SystemControl Callbacks

  void
  on_system_control_track_change (BmpSystemControl *control)
  {
    GtkListStore	    *playlist = bmp_playlist_get_playlist (bmp_playlist);
    GtkTreeRowReference     *reference;
    GtkTreePath             *path;
    GtkTreeIter              iter;
    char                    *artist, *album, *title;

    reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(playlist), "current"));
    if (!reference) 
    {
      bmp_status_icon_set_text (status_icon, ""); 
      return;
    }

    path = gtk_tree_row_reference_get_path (reference);
    if (!path) 
    {
      bmp_status_icon_set_text (status_icon, ""); 
      return;
    }

    gtk_tree_model_get_iter (GTK_TREE_MODEL (playlist), &iter, path);
    gtk_tree_path_free (path);

    gtk_tree_model_get (GTK_TREE_MODEL (playlist),
			&iter,
			0, &artist,
			1, &album,
			3, &title, -1);

    std::stringstream status_icon_text;
    status_icon_text << "<b>" << Glib::Markup::escape_text(artist ? artist : "") << "</b>\n" 
		     << "<i>" << Glib::Markup::escape_text(album ? album : "")  << "</i>\n" 
		     << Glib::Markup::escape_text (title ? title : "");
    free (artist);
    free (album);
    free (title);
    bmp_status_icon_set_text (status_icon, status_icon_text.str().c_str());
  }

  void
  set_ui_playstatus (Bmp::Playstatus playstatus)
  {
    using namespace Bmp;

    const char *stock_id;

      if (playstatus == PLAYSTATUS_PLAYING)
	{
            bmp_status_icon_set_from_pixbuf (status_icon,
                                             status_icon_image_playing);
	}
      else
      if (playstatus == PLAYSTATUS_PAUSED)
	{
            bmp_status_icon_set_from_pixbuf (status_icon,
                                             status_icon_image_paused);
	}
      else
	{
           bmp_status_icon_set_from_pixbuf  (status_icon,
                                             status_icon_image_default);
	}

      switch (playstatus)
      {
        case PLAYSTATUS_STOPPED: stock_id = GTK_STOCK_MEDIA_STOP; break;
        case PLAYSTATUS_PAUSED:  stock_id = GTK_STOCK_MEDIA_PAUSE; break;
        case PLAYSTATUS_PLAYING: stock_id = GTK_STOCK_MEDIA_PLAY; break;
        default: stock_id = GTK_STOCK_MEDIA_STOP; break;
      }

      bmp_status_icon_set_icon (status_icon, stock_id);

      if (playstatus == PLAYSTATUS_STOPPED)
	{
	  bmp_status_icon_set_text (status_icon, "");
	}
  }
  void
  on_system_control_set_playstatus (BmpSystemControl *system_control,
                                    Bmp::Playstatus   status)
  {
    set_ui_playstatus (status);
  }

  void
  set_ui_title    (const char  *title)
  {
    char *window_title;

    if (!title)
      {
        window_title = g_strdup_printf ("BMP");
      }
    else
      {
        window_title = g_strdup_printf ("BMP: %s", title);
      }

    gtk_window_set_title (GTK_WINDOW (Bmp::w_main->window), window_title);
    free (window_title);
  }

  void
  on_system_control_set_title (BmpSystemControl *control,
                               char             *title)
  {
    set_ui_title (title);
  }

  void
  on_system_control_app_busy (BmpSystemControl *control)
  {
    bmp_window_set_busy (Bmp::w_playlist->window);
    bmp_window_set_busy (Bmp::w_main->window);
  }

  void
  on_system_control_app_idle (BmpSystemControl *control)
  {
    bmp_window_set_idle (Bmp::w_playlist->window);
    bmp_window_set_idle (Bmp::w_main->window);
  }


  /*------------------------------------------------------------------------------------------------*/
  // MCS Callbacks

  void
  on_bmp_skin_changed (MCS_CB_DEFAULT_SIGNATURE)
  {
    std::string path = boost::get<std::string> (value);
    if (!bmp_winamp2_skin_load_skin (Bmp::winamp2_skin, path))
      {
        mcs->key_unset ("bmp", "skin");
      }
    bmp_window_main_configure (Bmp::w_main);
    bmp_window_playlist_configure (Bmp::w_playlist);
  }

  /*------------------------------------------------------------------------------------------------*/

  // Callbacks
  void
  action_play_files (GtkAction *action, gpointer data)
  {
    browser->run (Bmp::FileBrowser::PLAY);
  }

  void
  action_add_files  (GtkAction *action, gpointer data)
  {
    browser->run (Bmp::FileBrowser::ADD);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_about      (GtkAction *action, gpointer data)
  {
    about->present (); 
  }

  void
  action_library    (GtkAction *action, gpointer data)
  {
    library_ui->present (); 
  }

  void
  action_streams    (GtkAction *action, gpointer data)
  {
    streams->present (); 
  }

  void
  action_preferences
		    (GtkAction *action, gpointer data)
  {
    preferences->present (); 
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_iconify    (GtkAction *action, gpointer data)
  {
    if (mcs->key_get<bool>("bmp", "ui-esc-trayconify"))
    {
	Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					       (Bmp::actions_toggle->get_action (BMP_TOGGLE_ACTION_SHOW_UI));
	action->set_active (false);
    }
    else
    {
        gtk_window_iconify (GTK_WINDOW (Bmp::w_main->window));
    }
  }

  void
  action_quit       (GtkAction *action, gpointer data)
  {
    bmp_system_control_quit (bmp_system_control, NULL);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_remove     (GtkAction *action, gpointer data)
  {
    chroma_list_selection_remove_selected (CHROMA_LIST_SELECTION (c_tracklist));
    bmp_system_control_update_title (bmp_system_control);
  }

  void
  action_remove_all  (GtkAction *action, gpointer data)
  {
    bmp_playlist_tracklist_remove_rows_all (bmp_playlist);
    bmp_system_control_clear_playback_history (bmp_system_control);
    g_object_set (G_OBJECT (bmp_system_control), "current_index", int(-1), NULL);
    bmp_system_control_update_title (bmp_system_control);
  }

  void
  action_select_all (GtkAction *action, gpointer data)
  {
    chroma_list_selection_select_all (CHROMA_LIST_SELECTION (c_tracklist));
  }

  void
  action_clear_playback_history
		      (GtkAction *action, gpointer data)
  {
    bmp_system_control_clear_playback_history (bmp_system_control);
    gtk_action_set_sensitive (action, FALSE);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_export_tracklist
		    (GtkAction *action, gpointer data)
  {
      int response = export_dialog->run ();
      switch (response)
      {
	case GTK_RESPONSE_CANCEL: break;

	case GTK_RESPONSE_OK:
	  {
	    bmp_system_control_export_tracklist (bmp_system_control,
						 export_dialog->container.c_str(),
						 export_dialog->all_files, 
						 export_dialog->entry_filename->get_text().c_str(), 
						 NULL);

	  }
      }
      export_dialog->hide ();
  }

  void
  action_trackinfo  (GtkAction *action, gpointer data)
  {
    GtkListStore   *list;
    GtkTreeIter     iter;
    GtkTreePath	   *path;
    char	   *uri;

    path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION (c_tracklist));
    if (!path) return;

    list = bmp_playlist_get_playlist (bmp_playlist);
    gtk_tree_model_get_iter (GTK_TREE_MODEL(list), &iter, path);
    gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, COLUMN_URI, &uri, -1);
    gtk_tree_path_free (path);

    std::string uristd = uri;
    free (uri);
    trackinfo->show (uristd);
  }

  void
  action_trackinfo_clear
		     (GtkAction *action, gpointer data)
  {
    trackinfo->clear ();
  }

  void
  action_trackinfo_if_visible
		     (GtkAction *action, gpointer data)
  {
    if (trackinfo->is_visible ())
    {
      Bmp::actions->get_action (BMP_ACTION_FILEINFO)->activate ();
    }
  }

  /*------------------------------------------------------------------------------------------------*/

  bool
  get_cdinfo (std::string& cdindex, int& n_tracks)
  {
    MusicBrainz o;
    bool ret;

    o.UseUTF8 (true); 
    o.SetDevice (mcs->key_get<std::string>("audio", "cdrom-device"));

    ret = o.Query (MBQ_GetCDTOC);
    if (!ret) return false;

    cdindex  = o.Data (MBE_TOCGetCDIndexId);
    n_tracks = o.DataInt (MBE_TOCGetLastTrack);
    return true;
  }

  void
  action_add_cd (GtkAction *action, gpointer data)
  {
    int			n_tracks;
    std::string         cdindex;
    static boost::format cdda ("cdda://%s/%d");

    if (!get_cdinfo(cdindex, n_tracks)) return;

    char **uris = g_new0 (char*, n_tracks+1);
    library->cache_audiocd (cdindex);

    for (int n = 1; n <= n_tracks; n++)
      {
	uris[n-1] = strdup((cdda % cdindex % n).str().c_str()); 
      }

    uris[n_tracks] = 0;

    int out=0;
    bmp_system_control_add_uri_list (bmp_system_control,
				     (const gchar**)uris,
				     -1,
				     false,
				     false,
				     0,
				     &out,
				     NULL);
    g_strfreev (uris);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_prev   (GtkAction *action, gpointer data)
  {
    bmp_system_control_go_prev (bmp_system_control, NULL);
  }

  void 
  action_next   (GtkAction *action, gpointer data)
  {
    bmp_system_control_go_next (bmp_system_control, NULL);
  }

  void 
  action_play   (GtkAction *action, gpointer data)
  {
    bmp_system_control_play (bmp_system_control, NULL);
  }

  void 
  action_stop   (GtkAction *action, gpointer data)
  {
    bmp_system_control_stop (bmp_system_control, NULL);
  }
  
  void
  action_pause  (GtkAction *action, gpointer data)
  {
    bmp_system_control_pause (bmp_system_control, NULL);
  }

  /*------------------------------------------------------------------------------------------------*/

  void
  action_toggle_show_ui
		(GtkToggleAction *action,
				    gpointer data)
  {
    static bool visible = false;

    if (gtk_toggle_action_get_active (action))
      {
        if (x_main && y_main)
	{
	  gtk_window_move  (GTK_WINDOW (Bmp::w_main->window), x_main.get(), y_main.get());
	}
        gtk_window_present (GTK_WINDOW (Bmp::w_main->window));

        if (visible)
          {
            if (x_pl && y_pl)
	    {
	      gtk_window_move  (GTK_WINDOW (Bmp::w_playlist->window), x_pl.get(), y_pl.get());
	    }
            gtk_window_present (GTK_WINDOW (Bmp::w_playlist->window));
          }

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_main->window),
                                             false);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_main->window),
                                             false);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_playlist->window),
                                             false);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_playlist->window),
                                             false);
      }
    else
      {
	int _x_main, _y_main;
	int _x_pl,   _y_pl;

        gtk_window_get_position (GTK_WINDOW (Bmp::w_main->window),
				 &_x_main,
				 &_y_main);

        gtk_window_get_position (GTK_WINDOW (Bmp::w_playlist->window),
				 &_x_pl,
				 &_y_pl);

	x_pl = _x_pl;
	y_pl = _y_pl;

	x_main = _x_main;
	y_main = _y_main;

        gtk_widget_hide (GTK_WIDGET (Bmp::w_main->window));

	
        visible = GTK_WIDGET_VISIBLE (Bmp::w_playlist->window);
        if (visible)
          {
            gtk_widget_hide (Bmp::w_playlist->window);
          }

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_main->window),
                                             true);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_main->window),
                                             true);

        gtk_window_set_skip_taskbar_hint
                                        (GTK_WINDOW (Bmp::w_playlist->window),
                                             true);
        gtk_window_set_skip_pager_hint  (GTK_WINDOW (Bmp::w_playlist->window),
                                             true);
      }
  }

  void
  action_toggle_stop_after_current
		(GtkToggleAction *action,
				    gpointer data)
  {
    bmp_system_control_stop_after_current_set (bmp_system_control, gtk_toggle_action_get_active (action), NULL);
  }


  void
  action_toggle_shuffle
		(GtkToggleAction *action,
				    gpointer data)
  {
    bmp_system_control_shuffle_set (bmp_system_control, gtk_toggle_action_get_active (action), NULL);
  }

  void
  action_toggle_repeat
		(GtkToggleAction *action,
				    gpointer data)
  {
    bmp_system_control_repeat_set (bmp_system_control, gtk_toggle_action_get_active (action), NULL);
  }

  void
  action_toggle_keep_above
		(GtkToggleAction *action,
				    gpointer data)
  {
    bool keep_above = gtk_toggle_action_get_active (action);

    gtk_window_set_keep_above (GTK_WINDOW (Bmp::w_main->window), keep_above);
    gtk_window_set_keep_above (GTK_WINDOW (Bmp::w_playlist->window), keep_above);
  }

  void
  action_toggle_pl
		(GtkToggleAction *action,
				    gpointer data)
  {
    if (!gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action)))
      {
        int x, y, w, h;
        gtk_window_get_position (GTK_WINDOW (Bmp::w_playlist->window), &x, &y);
        gtk_window_get_size (GTK_WINDOW (Bmp::w_playlist->window), &w, &h);
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "x", gpointer (x));
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "y", gpointer (y));
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "w", gpointer (w));
        g_object_set_data (G_OBJECT (Bmp::w_playlist->window), "h", gpointer (h));
        gtk_widget_hide (Bmp::w_playlist->window);
      }
    else
      {
        int x = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "x"));
        int y = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "y"));
        int w = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "w"));
        int h = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "h"));

	gtk_window_resize (GTK_WINDOW (Bmp::w_playlist->window), w, h);

        if ((x > 0) && (y > 0))
          {
            gtk_window_move (GTK_WINDOW (Bmp::w_playlist->window), x, y);
          }
        gtk_widget_show_all (Bmp::w_playlist->window);
      }
  }

  /*------------------------------------------------------------------------------------------------*/

  // ActionEntries
  GtkActionEntry bmp_action_entries[] = {

      { "dummy", NULL, "dummy" },

      { BMP_ACTION_OPEN,
	    GTK_STOCK_MEDIA_PLAY,
	    N_("Play Files..."),
	    "<ctrl>o",
	    N_("Play Files..."),
	    G_CALLBACK (action_play_files)},

      { BMP_ACTION_ADD,
	    GTK_STOCK_ADD,
	    N_("Add Files..."),
	    "<ctrl>f",
	    NULL,
	    G_CALLBACK (action_add_files)},

      { BMP_ACTION_ADD_CD,
	    GTK_STOCK_CDROM,
	    N_("Add CD..."),
	    "<ctrl>d",
	    NULL,
	    G_CALLBACK (action_add_cd)},



      { BMP_ACTION_PREV,
	    GTK_STOCK_MEDIA_PREVIOUS,
	    N_("Previous"),
	    NULL,
	    N_("Previous"),
	    G_CALLBACK (action_prev)},

      { BMP_ACTION_PLAY,
	    GTK_STOCK_MEDIA_PLAY,
	    N_("Play"),
	    NULL,
	    N_("Play"),
	    G_CALLBACK (action_play)},

      { BMP_ACTION_PAUSE,
	    GTK_STOCK_MEDIA_PAUSE,
	    N_("Pause"),
	    "space",
	    N_("Pause"),
	    G_CALLBACK (action_pause)},

      { BMP_ACTION_STOP,
	    GTK_STOCK_MEDIA_STOP,
	    N_("Stop"),
	    NULL,
	    N_("Stop"),
	    G_CALLBACK (action_stop)},

      { BMP_ACTION_NEXT,
	    GTK_STOCK_MEDIA_NEXT,
	    N_("Next"),
	    NULL,
	    N_("Next"),
	    G_CALLBACK (action_next)},



      { BMP_ACTION_ICONIFY,
	    NULL,
	    N_("Minimize BMP"),
	    "<alt>m",
	    N_("Minimize"),
	    G_CALLBACK (action_iconify)},

      { BMP_ACTION_QUIT,
	    GTK_STOCK_QUIT,
	    N_("Quit BMP"),
	    "<control>w",
	    N_("Quit BMP"),
	    G_CALLBACK (action_quit)},

      { BMP_ACTION_QUIT_TITLE,
	    GTK_STOCK_QUIT,
	    "", "",
	    "",
	    G_CALLBACK (action_quit)},



      { BMP_ACTION_ABOUT,
	    BMP_STOCK_ABOUT,
	    N_("About BMP"),
	    NULL,
	    N_("About BMP"),
	    G_CALLBACK (action_about)},

      { BMP_ACTION_PREFERENCES,
	    BMP_STOCK_PREFERENCES,
	    N_("Preferences"),
	    "<ctrl>p",
	    N_("Preferences"),
	    G_CALLBACK (action_preferences)},

      { BMP_ACTION_LIBRARY,
	    BMP_STOCK_LIBRARY,
	    N_("Music Library"),
	    "<ctrl>l",
	    N_("Music Library"),
	    G_CALLBACK (action_library)},

      { BMP_ACTION_STREAMS,
	    BMP_STOCK_STREAMS,
	    N_("Radio Streams"),
	    "<ctrl>s",
	    N_("Radio Streams"),
	    G_CALLBACK (action_streams)},

      { BMP_ACTION_EXPORT,
	    GTK_STOCK_SAVE,
	    N_("Export Tracklist"),
	    "<ctrl>e",
	    NULL,
	    G_CALLBACK (action_export_tracklist)},

      { BMP_ACTION_FILEINFO,
	    GTK_STOCK_INFO,
	    N_("Track Information"),
	    "<ctrl>i",
	    NULL,
	    G_CALLBACK (action_trackinfo)},

      { BMP_ACTION_FILEINFO_CLEAR,
	    GTK_STOCK_DELETE,
	    N_("Clear Fileinfo"), 
	    NULL, 
	    NULL,
	    G_CALLBACK (action_trackinfo_clear)},

      { BMP_ACTION_FILEINFO_IF_VISIBLE,
	    GTK_STOCK_DELETE,
	    N_("Clear Fileinfo If Visible"), 
	    NULL, 
	    NULL,
	    G_CALLBACK (action_trackinfo_if_visible)},

    #if 0
      { BMP_ACTION_ADD_TO_NEW,
	    GTK_STOCK_ADD,
	    N_("Add Files to New..."),
	    NULL,	
	    NULL,
	    G_CALLBACK (NULL)},
    #endif

      { BMP_ACTION_DEL,
	    GTK_STOCK_DELETE,
	    N_("Remove Selected"),
	    "<ctrl>x",
	    NULL,
	    G_CALLBACK (action_remove)},

    #if 0
      { BMP_ACTION_CROP,
	    GTK_STOCK_CUT,
	    N_("Remove Unselected"),
	    NULL, 
	    NULL,
	    G_CALLBACK (NULL)},
    #endif

      { BMP_ACTION_DEL_ALL,
	    GTK_STOCK_CLEAR,
	    N_("Remove All"),
	    NULL,
	    NULL,
	    G_CALLBACK (action_remove_all)},

      { BMP_ACTION_CLEAR_PLAYBACK_HISTORY,
	    GTK_STOCK_UNDO,
	    N_("Clear Playback History"),
	    "<ctrl>h",
	    NULL,
	    G_CALLBACK (action_clear_playback_history)},

    #if 0
      { BMP_ACTION_JTT,
	    GTK_STOCK_JUMP_TO,
	    N_("Jump To Track"),
	    "<ctrl>j",
	    NULL,
	    G_CALLBACK (NULL)},
    #endif

      { BMP_ACTION_SELECT_ALL,
	    NULL,
	    N_("Select All"),
	    "<control>a",
	    NULL,
	    G_CALLBACK (action_select_all) },

    #if 0
      { BMP_ACTION_SELECT_NONE,
	    NULL,
	    N_("Select None"),
	    "<control><shift>a",
	    NULL,
	    G_CALLBACK (NULL) },

      { BMP_ACTION_SELECT_INVERT,
	    NULL,
	    N_("Invert Selection"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_VISIBLE,
	    NULL,
	    N_("Sort by Visible Title"),
	    "<shift>s",
	    NULL,
	    G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_ARTIST,
	    NULL,
	    N_("Sort by Artist"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_ALBUM,
	    NULL,
	    N_("Sort by Album"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_TITLE,
	    NULL,
	    N_("Sort by Title"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_GENRE,
	    NULL,
	    N_("Sort by Genre"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_TRACK,
	    NULL,
	    N_("Sort by Tracknumber"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL)},

      { BMP_ACTION_SORT_LOCATION,
	    NULL,
	    N_("Sort by Location"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL)}
    #endif

  };

  GtkToggleActionEntry bmp_toggle_action_entries[] = {

      { BMP_TOGGLE_ACTION_FOLLOW_CURRENT,
	    NULL,
	    N_("Follow Currently Played Track"),
	    NULL,
	    NULL,
	    G_CALLBACK (NULL),
	    FALSE },

      { BMP_TOGGLE_ACTION_STOP_AFTER_CURRENT,
	    NULL,
	    N_("Stop After Current Track"),
	    NULL,
	    NULL,
	    G_CALLBACK (action_toggle_stop_after_current),
	    FALSE },

      { BMP_TOGGLE_ACTION_SHUFFLE,
	    NULL,
	    N_("_Shuffle"),
	    NULL,
	    N_("Switch between random and continuous playback"),
	    G_CALLBACK (action_toggle_shuffle),
	    FALSE },

      { BMP_TOGGLE_ACTION_REPEAT,
	    NULL,
	    N_("_Repeat"),
	    NULL,
	    N_("Whether to repeat the tracklist or not"),
	    G_CALLBACK (action_toggle_repeat),
	    FALSE },

      { BMP_TOGGLE_ACTION_PL,
	    NULL,
	    N_("Show Playlist Window"),
	    NULL,
	    N_("Toggles playlist window visibility"),
	    G_CALLBACK (action_toggle_pl),
	    FALSE },

      { BMP_TOGGLE_ACTION_SCROLL,
	    NULL,
	    N_("Scroll Textbox"),
	    NULL,
	    N_("Toggles textbox scroll"),
	    G_CALLBACK (NULL),
	    TRUE },

      { BMP_TOGGLE_ACTION_SHOW_UI,
	    NULL,
	    N_("Show UI"),
	    NULL,
	    N_("Toggle BMP UI Visiblity"),
	    G_CALLBACK (action_toggle_show_ui),
	    TRUE },

      { BMP_TOGGLE_ACTION_KEEP_ABOVE,
	    NULL,
	    N_("Keep BMP on top"),
	    NULL,
	    N_("Keep BMP on top"),
	    G_CALLBACK (action_toggle_keep_above),
	    FALSE },
  };

  void
  register_stock_icons ()
  {
    struct
    {
      char* filename;
      char* stock_id;
    }
      list[] = {
    { "library.png",	  BMP_STOCK_LIBRARY	},
    { "streams.png",	  BMP_STOCK_STREAMS	},
    { "bmp.png",	  BMP_STOCK_ABOUT	},
    { "preferences.png",  BMP_STOCK_PREFERENCES },
    { "bookmarks.png",	  BMP_STOCK_BOOKMARKS   },
    { "xiph.png",	  BMP_STOCK_XIPH        },
    { "shoutcast.png",	  BMP_STOCK_SHOUTCAST   },
    };

    Glib::RefPtr<Gtk::IconFactory> factory = Gtk::IconFactory::create ();
    for (unsigned int n = 0; n < G_N_ELEMENTS (list); ++n)
    {
      std::string filename = Glib::build_filename (DATA_DIR G_DIR_SEPARATOR_S "images" G_DIR_SEPARATOR_S "stock", list[n].filename);
      Glib::RefPtr<Gdk::Pixbuf> pixbuf = Gdk::Pixbuf::create_from_file (filename);
      factory->add (Gtk::StockID (list[n].stock_id), Gtk::IconSet(pixbuf));
    }

    factory->add_default ();
  }

} // ns::<unnamed>

namespace Bmp
{
  void
  raise_windows ()
  {
    Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					     (actions_toggle->get_action (BMP_TOGGLE_ACTION_SHOW_UI));
    action->set_active (true);
    gtk_window_present (GTK_WINDOW (Bmp::w_playlist->window));
    gtk_window_present (GTK_WINDOW (Bmp::w_main->window));
  }

  void
  display_windows ()
  {
    int x,y,w,h;
    bool visible;

    w = mcs->key_get<int>("playlist-window", "width");
    h = mcs->key_get<int>("playlist-window", "height");
    x = mcs->key_get<int>("playlist-window", "pos_x");
    y = mcs->key_get<int>("playlist-window", "pos_y");

    if ((x < 0) || (y < 0)) 
      {
        x = gdk_screen_get_width  (gdk_screen_get_default());
        y = gdk_screen_get_height (gdk_screen_get_default());
        x /= 2;
        y /= 2;
        x -= BMP_WINDOW_MAIN_WIDTH;
        y -= BMP_WINDOW_MAIN_HEIGHT;
      }

    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "x", gpointer (x));
    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "y", gpointer (y));
    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "w", gpointer (w));
    g_object_set_data (G_OBJECT   (Bmp::w_playlist->window), "h", gpointer (h));

    visible = mcs->key_get<bool>("playlist-window", "visible");

    if (visible)
    {
        Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					       (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL));
	action->set_active (true);
    }

    x = mcs->key_get<int>("main-window", "pos_x");
    y = mcs->key_get<int>("main-window", "pos_y");

    if ((x < 0) || (y < 0)) 
    {
        x = gdk_screen_get_width (gdk_screen_get_default());
        y = gdk_screen_get_height (gdk_screen_get_default());
        x /= 2;
        y /= 2;
        x -= BMP_WINDOW_MAIN_WIDTH/2;
        y -= BMP_WINDOW_MAIN_HEIGHT/2;

    }

    gtk_window_move   (GTK_WINDOW (Bmp::w_main->window), x, y);
    g_object_set_data (G_OBJECT   (Bmp::w_main->window), "x", gpointer (x));
    g_object_set_data (G_OBJECT   (Bmp::w_main->window), "y", gpointer (y));

    gtk_widget_show_all (Bmp::w_main->window);
  }

  void
  shell_init () 
  {
      GtkUIManager    *_ui_manager = gtk_ui_manager_new ();
      GtkActionGroup  *_actions;
      GtkActionGroup  *_actions_toggle;

      register_stock_icons ();

      _actions = gtk_action_group_new (BMP_ACTION_GROUP);
      _actions_toggle = gtk_action_group_new (BMP_ACTION_GROUP_TOGGLE);

      gtk_action_group_set_translate_func (_actions,
                bmp_menu_translate,
                NULL,
                NULL);

      gtk_action_group_set_translate_func (_actions_toggle,
                bmp_menu_translate,
                NULL,
                NULL);
      
      gtk_ui_manager_insert_action_group (_ui_manager, _actions, 0);
      gtk_ui_manager_insert_action_group (_ui_manager, _actions_toggle, 0);
      gtk_action_group_set_sensitive (_actions, true);
      gtk_action_group_set_sensitive (_actions_toggle, true);

      gtk_action_group_add_actions	  (_actions,
					   bmp_action_entries,
					   G_N_ELEMENTS (bmp_action_entries),
					   0);

      gtk_action_group_add_toggle_actions (_actions_toggle,
					   bmp_toggle_action_entries,
					   G_N_ELEMENTS (bmp_toggle_action_entries),
					   0);

      gtk_ui_manager_add_ui_from_file (_ui_manager, DATA_DIR G_DIR_SEPARATOR_S BMP_BASENAME_UI_DIR G_DIR_SEPARATOR_S "main.ui",	  NULL);
      gtk_ui_manager_add_ui_from_file (_ui_manager, DATA_DIR G_DIR_SEPARATOR_S BMP_BASENAME_UI_DIR G_DIR_SEPARATOR_S "trayicon.ui",	  NULL);
      gtk_ui_manager_add_ui_from_file (_ui_manager, DATA_DIR G_DIR_SEPARATOR_S BMP_BASENAME_UI_DIR G_DIR_SEPARATOR_S "tracklist.ui",  NULL);

      ui_manager = Glib::wrap (_ui_manager);
      actions = Glib::wrap (_actions);
      actions_toggle = Glib::wrap (_actions_toggle);

      // Set up internet dependent actions 
      if (!Bmp::Network::is_connected ())
      {
	actions->get_action (BMP_ACTION_STREAMS)->set_sensitive (false);
      }

      // ResourceManager paths
      bmp_resource_manager_create_path (bmp_rm, NULL, "tooltips"); 

      // Init C++ objects needed for Winamp2.x skinning engine
      mcs_bind	    = new Mcs::Bind (mcs);

      // Init Winamp2.x skinning system
      winamp2_skin  = bmp_winamp2_skin_new ();
      bmp_winamp2_skin_load_skin (winamp2_skin, mcs->key_get<std::string>("bmp","skin")); 
      w_main  = bmp_window_main_new ();
      w_playlist  = bmp_window_playlist_new ();
      mcs->subscribe ("Shell", "bmp", "skin", sigc::ptr_fun (&on_bmp_skin_changed));

      // Init C++ objects, the rest 
      preferences   = Bmp::Preferences::create ();
      browser	    = Bmp::FileBrowser::create ();
      trackinfo	    = Bmp::TrackInfo::create ();
      if (Bmp::Network::is_connected () )
      {
	streams	    = Bmp::Streams::Dialog::create ();
      }
      library_ui    = Bmp::LibraryUI::create ();
      export_dialog = Bmp::ExportDialog::create ();
      about	    = new Bmp::AboutDialog;

      // Tray icon
      status_icon_image_default =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-default.png",
                                  NULL);
      status_icon_image_paused =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-paused.png",
                                  NULL);
      status_icon_image_playing =
        gdk_pixbuf_new_from_file (DATA_DIR G_DIR_SEPARATOR_S "icons/tray-icons/tray-icon-playing.png",
                                  NULL);
      status_icon =
        bmp_status_icon_new_from_pixbuf (status_icon_image_default);

      bmp_status_icon_set_tooltip (status_icon, "", GTK_STOCK_MEDIA_STOP);
      bmp_status_icon_set_visible (status_icon, TRUE);

      g_object_connect (G_OBJECT (status_icon),

			    "signal::click",
			    G_CALLBACK(status_icon_click),
			    NULL,

			    "signal::popup-menu",
			    G_CALLBACK(status_icon_popup_menu),
			    NULL,

			    "signal::scroll-up",
			    G_CALLBACK(status_icon_scroll_up),
			    NULL,

			    "signal::scroll-down",
			    G_CALLBACK(status_icon_scroll_down),
			    NULL, NULL);


      // SystemControl callbacks
      g_object_connect (bmp_system_control,

			"signal::set-playstatus",
			G_CALLBACK (on_system_control_set_playstatus),
			NULL,

			"signal::set-title",
			G_CALLBACK (on_system_control_set_title),
			NULL,

			"signal::track-change",
			G_CALLBACK (on_system_control_track_change),
			NULL,

			"signal::app-busy",
			G_CALLBACK (on_system_control_app_busy),
			NULL,

			"signal::app-idle",
			G_CALLBACK (on_system_control_app_idle),
			NULL,
  
			NULL);

    display_windows ();

    // MCS Bindings
    Glib::RefPtr<Gtk::ToggleAction> action;
    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					       (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL));
    mcs_bind->bind_toggle_action (action, "playlist-window", "visible");

    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					       (actions_toggle->get_action (BMP_TOGGLE_ACTION_KEEP_ABOVE));
    mcs_bind->bind_toggle_action (action, "bmp", "keep-above");

    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					       (actions_toggle->get_action (BMP_TOGGLE_ACTION_REPEAT));
    mcs_bind->bind_toggle_action (action, "bmp", "repeat");

    action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
					       (actions_toggle->get_action (BMP_TOGGLE_ACTION_SHUFFLE));
    mcs_bind->bind_toggle_action (action, "bmp", "shuffle");


    set_ui_playstatus (Bmp::Playstatus (play->property_status().get_value()));

    char *title; 
    g_object_get (G_OBJECT (bmp_system_control), "current_title", &title, NULL);
    set_ui_title (title);
    free (title);

    running = true;
    g_message ("Running: Shell");
  }

  void
  shell_stop ()
  {
    int _x_main, _y_main;

    gtk_window_get_position (GTK_WINDOW (Bmp::w_main->window),
			     &_x_main,
			     &_y_main);

    mcs->key_set<int>("main-window", "pos_x", _x_main);
    mcs->key_set<int>("main-window", "pos_y", _y_main);

    int x, y, w, h;

    if (GTK_WIDGET_VISIBLE (Bmp::w_playlist->window))
      {
	gtk_window_get_position (GTK_WINDOW (Bmp::w_playlist->window),
				      &x,
				      &y);

	gtk_window_get_size     (GTK_WINDOW (Bmp::w_playlist->window),
				      &w,
				      &h);

      }
    else 
      {
        x = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "x"));
        y = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "y"));
        w = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "w"));
        h = int (g_object_get_data (G_OBJECT (Bmp::w_playlist->window), "h"));
      }

    mcs->key_set<int>("playlist-window", "width",  w);
    mcs->key_set<int>("playlist-window", "height", h);
    mcs->key_set<int>("playlist-window", "pos_x",  x);
    mcs->key_set<int>("playlist-window", "pos_y",  y);

    // Dispose C Objects
    g_object_unref (Bmp::w_main);
    g_object_unref (Bmp::w_playlist);
    g_object_unref (Bmp::winamp2_skin);

    // Delete C++ objects 
    delete  export_dialog;
    delete  about;
    delete  browser;
    delete  trackinfo;
    delete  streams;
    delete  library_ui;
    delete  preferences;
    delete  mcs_bind;

    running = false;

    g_message ("Stopped: Shell");
  }

  bool
  shell_running ()
  {
    return running;
  }
}
