#ifndef BMP_TRACKLIST_CLASS_HPP
#define BMP_TRACKLIST_CLASS_HPP

#include <list>
#include <vector>
#include <map>
#include <string>
#include <glibmm.h>
#include <gtkmm.h>

#include <boost/optional.hpp>

#include <chroma/chroma-list.h>
#include <chroma/chroma-list-selection.h>

#include "database.hpp"
#include "uri++.hpp"
#include "file_utils.hpp"

namespace Bmp
{
  /** Row GUID
   */
  typedef unsigned long GUID;

  /** Map mapping a GUID <-> TreeModel::iterator
   */
  typedef std::map<GUID, Gtk::TreeModel::iterator> GUIDMap;
  typedef std::list<Gtk::TreeModel::iterator> IterList;


  class PlaybackHistory
  {
  public:

      PlaybackHistory ();
      ~PlaybackHistory ();

      /** Rewinds the history to the beginning.
       *
       */
      void
      rewind ();

      /** Truncates the end of the history and appends GUID guid
       *
       * @param guid The GUID to append
       *
       */
      void
      set (GUID guid);

      /** Prepends GUID guid to the history
       *
       * @param guid The GUID to prepend
       *
       */
      void
      prepend (GUID guid);


      /** Appends GUID guid to the history
       *
       * @param guid The GUID to prepend
       *
       */
      void
      append (GUID guid);

      /** Goes one step back in the history
       *
       */
      void
      back ();

      /** Goes one step forward in the history
       *
       */
      void
      forward ();

      /** Clears the history
       *
       */
      void
      clear ();

      /** Removes the given GUID from the history
       *
       * @param guid The GUID to remove
       *
       */
      void
      remove (GUID guid);

      /** Returns the GUID of the current track
       */
      GUID
      get_mark () const
      {
          return (*mark);
      }

      /** Returns true if the history size is > 0
       */
      operator bool ()
      {
          return history.size ();
      }

  private:

      typedef std::list<GUID> History;

      Glib::Mutex       lock;
      History           history;
      History::iterator mark;
  };

  class TracklistColumns
      : public Gtk::TreeModel::ColumnRecord
  {
  public:

      Gtk::TreeModelColumn<Glib::ustring> artist;
      Gtk::TreeModelColumn<Glib::ustring> album;
      Gtk::TreeModelColumn<int>           track;
      Gtk::TreeModelColumn<Glib::ustring> title;
      Gtk::TreeModelColumn<Glib::ustring> genre;
      Gtk::TreeModelColumn<int>           date;
      Gtk::TreeModelColumn<int>           time;
      Gtk::TreeModelColumn<GUID>	  guid;
      Gtk::TreeModelColumn<URI>           uri;

      TracklistColumns ()
      {
          add (artist);
          add (album);
          add (track);
          add (title);
          add (genre);
          add (date);
          add (time);
          add (guid);
          add (uri);
      }
  };

  class Tracklist
  {

    public:

      Tracklist ();
      ~Tracklist ();

      TracklistColumns tracklist_columns;

      typedef sigc::signal<void, std::vector<int> > SignalRemoved;
      typedef sigc::signal<void, std::vector<int> > SignalAdded;

      SignalRemoved&
      signal_removed ()
      {
	return signal_removed_;
      }

      SignalAdded&
      signal_added ()
      {
	return signal_added_;
      }

      /** Insert rows at the given index into the playlist
       *
       * @param uris list of URIs to insert into the playlist
       * @param index The index at which to insert the items
       * @returns The new end of the list
       *
       */
      unsigned int
      insert_rows (const Util::FileList      &uris,
                   const unsigned int	     index);

      GtkListStore*
      c_liststore ()
      {
	return Glib::unwrap(playlist);
      }

      Glib::RefPtr<Gtk::ListStore>
      liststore ()
      {
	return playlist;
      }

      //Stuff
#if 0 
      bool 
      get_iter_by_guid (GUID guid, Gtk::TreeModel::iterator& iter)
      {
	GUIDMap::iterator i = mapping.find (guid);
	if (i == mapping.end ()) return false; 
	iter = i->second;
      }

      bool
      get_row_by_guid (GUID guid, int& row)
      {
	Gtk::TreeModel::iterator iter;
	if (!get_iter_by_guid (guid, iter)) return false;
	Gtk::TreePath path (iter);
	GtkTreePath *path_C = path.gobj ();
	row = gtk_tree_path_get_indices (path_C)[0];
	return true;
      }

      bool
      get_current_iter (Gtk::TreeModel::iterator& iter)
      {
	Gtk::TreeModel::iterator iter_;
	if (!current) return false;
	if (!get_iter_by_guid (current, iter_)) return false;
	iter = iter_;
	return true;
      }


      std::string
      get_title_nth (int index);
#endif

      /** This function is needed e.g. when removing rows. We also need to remove them from history.
       *
       */
      GUID
      get_guid_by_iter (const Gtk::TreeModel::iterator& iter)
      {
	return (*iter)[tracklist_columns.guid];
      }

      void
      path_remove (GtkTreePath *path);

    private:

      void
      row_inserted (const Gtk::TreeModel::Path     &path,
                    const Gtk::TreeModel::iterator &iter);

      Glib::RefPtr<Gtk::ListStore>  playlist;
      boost::optional<GUID>	    current;
      GUIDMap			    mapping;
      PlaybackHistory		    history;
      static unsigned long          guid_base;

      SignalRemoved		    signal_removed_;
      SignalAdded		    signal_added_;
  };

} // Bmp namespace

#endif // BMP_TRACKLIST_CLASS_HPP
